# Licensed under the GPLv3 - see LICENSE
import pytest
import numpy as np
from astropy import units as u
from astropy.time import Time

from ... import mark4


class TestCorruptFile:
    def setup_class(cls):
        time = Time('2010-11-12T13:14:15')
        cls.nchan = 2
        cls.header0 = mark4.Mark4Header.fromvalues(
            time=time, ntrack=16, nchan=cls.nchan, fanout=4)
        cls.data = np.zeros((2*cls.header0.frame_nbytes, 2))
        cls.data.reshape(-1, 4, 2)[160:] = [
            [-1, 1], [-3, 3], [1, -1], [3, -3]]
        cls.sample_rate = 100 * u.kHz
        cls.kwargs = dict(sample_rate=cls.sample_rate,
                          ref_time=time)

    def fake_file(self, tmpdir, nframes=8):
        filename = str(tmpdir.join('fake.mark4'))
        with mark4.open(filename, 'ws', header0=self.header0,
                        sample_rate=self.sample_rate) as fw:
            for _ in range(nframes):
                fw.write(self.data)
        return filename

    def test_fake_file(self, tmpdir):
        fake_file = self.fake_file(tmpdir)
        with mark4.open(fake_file, 'rs', **self.kwargs) as fr:
            data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        assert np.all(data.astype(int) == self.data)

    def corrupt_copy(self, filename, missing):
        corrupt_name = filename.replace('.mark4', '_corrupt.mark4')
        with open(filename, 'rb') as fr, \
                open(corrupt_name, 'wb') as fw:
            fw.write(fr.read(missing.start))
            fr.seek(missing.stop)
            fw.write(fr.read())
        return corrupt_name

    @pytest.mark.parametrize('frame_nr', [1, 3, slice(3, 5)])
    def test_missing_frames(self, frame_nr, tmpdir):
        if not isinstance(frame_nr, slice):
            frame_nr = slice(frame_nr, frame_nr+1)
        missing = slice(frame_nr.start * self.header0.frame_nbytes,
                        frame_nr.stop * self.header0.frame_nbytes)
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, missing)

        # Check that bad frames are found with verify only.
        with mark4.open(corrupt_file, 'rs', verify=True, **self.kwargs) as fv:
            assert not fv.info.readable
            assert not fv.info.checks['continuous']
            assert 'continuous' in fv.info.errors
            expected_msg = 'While reading at {}'.format(
                frame_nr.start * fv.samples_per_frame)
            assert expected_msg in fv.info.errors['continuous']

        # While only warnings are given when it is fixable.
        with mark4.open(corrupt_file, 'rs', verify='fix', **self.kwargs) as ff:
            assert ff.info.readable
            assert 'fixable' in ff.info.checks['continuous']
            assert 'continuous' in ff.info.warnings
            assert expected_msg in ff.info.warnings['continuous']
            assert 'problem loading frame' in ff.info.warnings['continuous']

        with mark4.open(corrupt_file, 'rs', **self.kwargs) as fr:
            with pytest.warns(UserWarning, match='problem loading'):
                data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        assert np.all(data[:frame_nr.start].astype(int) == self.data)
        assert np.all(data[frame_nr.stop:].astype(int) == self.data)
        assert np.all(data[frame_nr] == 0.)

    @pytest.mark.parametrize('missing_bytes', [
        slice(0, 40000),  # Remove whole last frame set.
        slice(0, 320),  # Remove header of last frame.
        slice(8, 16),  # Corrupt header of last frame.
        slice(0, 1),  # Corrupt header byte of last frame.
        slice(10, 11),  # Corrupt header byte of last frame.
        slice(319, 320),  # Corrupt header byte of last frame.
        slice(400, 401),  # Corrupt payload byte of last frame.
        slice(39999, 40000),  # last byte of last frame.
    ])
    def test_missing_end(self, missing_bytes, tmpdir):
        # In all these cases, the data read should just be short, but
        # it varies by how much: If the sync pattern is still there,
        # or the whole last frame is missing, we'll lose the last
        # frame.  Otherwise, the one-but-last header will not be
        # considered OK and we lose that frame too.
        if missing_bytes.start > 192 or (missing_bytes.start == 0
                                         and missing_bytes.stop == 40000):
            expected = 7
        else:
            expected = 6
        missing = slice(missing_bytes.start + 7*self.header0.frame_nbytes,
                        missing_bytes.stop + 7*self.header0.frame_nbytes)
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, missing)
        with mark4.open(corrupt_file, 'rs', **self.kwargs) as fr:
            assert len(fr.fh_raw.locate_frames(
                fr.header0, maximum=1000000)) == expected
            assert fr.size == expected * self.data.size
            data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        assert np.all(data.astype(int) == self.data)

    @pytest.mark.parametrize('missing_bytes,missing_frames', [
        (slice(40000, 80000), slice(1, 2)),  # Remove frame 1.
        (slice(78000, 82000), slice(1, 3)),  # Corrupt frames 1,2.
        (slice(80010, 80100), slice(1, 3)),  # Corrupted header 2.
    ])
    def test_missing_middle(self, missing_bytes, missing_frames, tmpdir):
        # In all these cases, some data will be missing.
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, missing_bytes)
        # Check that bad frames are found with verify only.
        with mark4.open(corrupt_file, 'rs', verify=True, **self.kwargs) as fv:
            assert not fv.info.readable
            assert not fv.info.checks['continuous']
            assert 'continuous' in fv.info.errors
            expected_msg = 'While reading at {}'.format(
                missing_frames.start * fv.samples_per_frame)
            assert expected_msg in fv.info.errors['continuous']

        with mark4.open(corrupt_file, 'rs', **self.kwargs) as fr:
            assert fr.size == 8 * self.data.size
            with pytest.warns(UserWarning, match='problem loading frame'):
                data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        expected = np.stack([self.data] * 8)
        expected[missing_frames] = 0.
        assert np.all(data.astype(int) == expected)

    def test_duplicate_data(self, tmpdir):
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, slice(100000, 40000))
        with mark4.open(corrupt_file, 'rs', verify='fix', **self.kwargs) as fv:
            with pytest.raises(Exception, match='excess data'):
                fv.read()



# Licensed under the GPLv3 - see LICENSE
import os
import copy
import pickle

import pytest
import numpy as np
import astropy.units as u
from astropy.time import Time

from ... import gsb
from ..payload import decode_4bit, encode_4bit
from ...data import (SAMPLE_GSB_RAWDUMP_HEADER as SAMPLE_RAWDUMP_HEADER,
                     SAMPLE_GSB_RAWDUMP as SAMPLE_RAWDUMP,
                     SAMPLE_GSB_PHASED_HEADER as SAMPLE_PHASED_HEADER,
                     SAMPLE_GSB_PHASED as SAMPLE_PHASED)


class TestGSB:

    def setup(self):
        # For all sample files, each frame spans 0.25165824 sec.
        self.frame_rate = (1e8 / 3) / 2**23 * u.Hz
        # Payload size for all sample files is 2**12 bytes.
        self.payload_nbytes = 2**12

    def test_rawdump_header(self):
        with open(SAMPLE_RAWDUMP_HEADER, 'rt') as fh:
            header = gsb.GSBHeader.fromfile(fh, verify=True)
        assert header.mode == 'rawdump'
        assert header['gps'] == '2015 04 27 18 45 00 0.000000240'
        # Includes UTC offset.
        assert abs(header.time
                   - Time('2015-04-27T13:15:00.000000240')) < 1.*u.ns
        header2 = gsb.GSBHeader.fromkeys(**header)
        assert header2 == header
        header3 = gsb.GSBHeader.fromvalues(mode='rawdump', **header2)
        assert header3 == header2
        assert header3.nbytes == header2.nbytes
        with pytest.raises(TypeError):
            gsb.GSBHeader.fromvalues(**header)
        with pytest.raises(TypeError):
            gsb.GSBHeader(None)
        # Check that recovering with the actual header type works as well.
        header4 = type(header).fromkeys(**header)
        assert header4 == header
        # Check that trying to initialize a phased header doesn't work.
        with pytest.raises(KeyError):
            gsb.header.GSBPhasedHeader.fromkeys(**header)
        header5 = header.copy()
        assert header5 == header

    def test_phased_header(self, tmpdir):
        with open(SAMPLE_PHASED_HEADER, 'rt') as fh:
            header = gsb.GSBHeader.fromfile(fh, verify=True)
            fh.seek(0)
            h_raw = fh.readline().strip()
        assert header.mode == 'phased'
        assert header['pc'] == h_raw[:28]
        assert header['gps'] == h_raw[29:60]
        assert header['seq_nr'] == 9995
        assert header['mem_block'] == 3
        assert abs(header.pc_time
                   - Time('2013-07-27T21:23:55.517535')) < 1.*u.ns
        assert header.gps_time == header.time
        assert abs(header.time
                   - Time('2013-07-27T21:23:55.3241088')) < 1.*u.ns
        assert header.mutable is False
        with pytest.raises(TypeError):
            header['mem_block'] = 0

        with open(str(tmpdir.join('test.timestamp')), 'w+t') as s:
            header.tofile(s)
            s.seek(0)
            assert s.readline().strip() == h_raw
            s.seek(0)
            header2 = gsb.GSBHeader.fromfile(s)
            with pytest.raises(EOFError):
                gsb.GSBHeader.fromfile(s)

        assert header == header2
        assert header2.mutable is False
        header3 = gsb.GSBHeader.fromkeys(**header)
        assert header3 == header
        assert header3.mutable is True
        with pytest.raises(KeyError):
            gsb.GSBHeader.fromkeys(extra=1, **header)
        with pytest.raises(KeyError):
            kwargs = dict(header)
            kwargs.pop('seq_nr')
            gsb.GSBHeader.fromkeys(**kwargs)
        # Try initialising with properties instead of keywords.
        header4 = gsb.GSBHeader.fromvalues(time=header.time,
                                           pc_time=header.pc_time,
                                           seq_nr=header['seq_nr'],
                                           mem_block=header['mem_block'])
        assert header4 == header
        assert header4.mutable is True
        header5 = header.copy()
        assert header5 == header
        assert header5.mutable is True
        header5['seq_nr'] = header['seq_nr'] + 1
        assert header5['seq_nr'] == header['seq_nr'] + 1
        assert header5 != header
        header5.time = Time('2014-01-20T05:30:00')
        assert header5['gps'] == '2014 01 20 11 00 00 0.000000000'
        header5['gps'] = '2014 01 20 11 00 00.000000000 0'
        with pytest.raises(ValueError):
            header5.time

    def test_rawdump_header_seek_offset(self):
        fh = open(SAMPLE_RAWDUMP_HEADER, 'rt')

        header = gsb.GSBHeader.fromfile(fh, verify=True)
        # Includes 1 trailing blank space, one line separator
        header_nbytes = header.nbytes
        assert (header_nbytes
                == len(' '.join(header.words) + ' ' + os.linesep))
        assert header.seek_offset(1) == header_nbytes
        assert header.seek_offset(12) == 12 * header_nbytes

        # Note that text pointers can't seek from current position.
        # Seek 2nd header.
        fh.seek(header.seek_offset(1))
        header1 = gsb.GSBHeader.fromfile(fh, verify=True)
        assert abs(header1.time
                   - Time('2015-04-27T13:15:00.251658480')) < 1.*u.ns

        fh.seek(header.seek_offset(9))
        header2 = gsb.GSBHeader.fromfile(fh, verify=True)
        assert abs(header2.time
                   - Time('2015-04-27T13:15:02.264924400')) < 1.*u.ns

        fh.close()

    def test_phased_header_seek_offset(self):
        fh = open(SAMPLE_PHASED_HEADER, 'rt')
        header1 = gsb.GSBHeader.fromfile(fh, verify=True)
        fh.seek(0)
        header_nbytes = header1.nbytes
        assert header_nbytes == header1.seek_offset(1)
        # Check that seek_offset is working properly.
        n_1000_0 = 1000 - 9995
        offset_to_1000_0 = header1.seek_offset(n_1000_0)
        assert offset_to_1000_0 == n_1000_0 * header_nbytes
        # Go to 1000.
        assert (header1.seek_offset(n_1000_0 - 1)
                == (n_1000_0 - 1) * header_nbytes + 1)
        # Go to 100 (header decreases by 1 chr).
        assert (header1.seek_offset(n_1000_0 - 900)
                == (n_1000_0 - 900) * header_nbytes + 900)
        # Go to 99 (header decreases by 2 chr).
        assert (header1.seek_offset(n_1000_0 - 901)
                == (n_1000_0 - 901) * header_nbytes + 902)
        # Go to 100001.
        assert (header1.seek_offset(n_1000_0 + 99001)
                == (n_1000_0 + 99001) * header_nbytes + 90002)

        # Try retrieving headers using seek_offset.
        fh.seek(header1.seek_offset(3))
        header2 = gsb.GSBHeader.fromfile(fh, verify=True)
        fh.seek(3 * header_nbytes)
        h2_raw = fh.readline().strip()
        assert header2.mode == 'phased'
        assert header2['pc'] == h2_raw[:28]
        assert header2['gps'] == h2_raw[29:60]
        assert header2['seq_nr'] == 9998
        assert header2['mem_block'] == 6
        assert abs(header2.pc_time
                   - Time('2013-07-27T21:23:56.272643')) < 1.*u.ns
        assert header2.gps_time == header2.time
        assert abs(header2.time
                   - Time('2013-07-27T21:23:56.079083520')) < 1.*u.ns

        # Retrieve beyond sequence number 10000.
        fh.seek(header1.seek_offset(8))
        header3 = gsb.GSBHeader.fromfile(fh, verify=True)
        fh.seek(8 * header_nbytes + 3)
        h3_raw = fh.readline().strip()
        assert header3.mode == 'phased'
        assert header3['pc'] == h3_raw[:28]
        assert header3['gps'] == h3_raw[29:60]
        assert header3['seq_nr'] == 10003
        assert header3['mem_block'] == 3
        assert abs(header3.pc_time
                   - Time('2013-07-27T21:23:57.530805')) < 1.*u.ns
        assert header3.gps_time == header3.time
        assert abs(header3.time
                   - Time('2013-07-27T21:23:57.337374720')) < 1.*u.ns

        fh.close()

    def test_header_non_gmrt(self):
        # Assume file is non-GMRT.
        with open(SAMPLE_PHASED_HEADER, 'rt') as fh:
            header = gsb.GSBHeader.fromfile(fh, verify=True,
                                            utc_offset=0.*u.hr)
        assert abs(header.pc_time
                   - Time('2013-07-28T02:53:55.517535')) < 1.*u.ns
        assert header.gps_time == header.time
        assert abs(header.time
                   - Time('2013-07-28T02:53:55.3241088')) < 1.*u.ns

    @pytest.mark.parametrize('sample', [
        SAMPLE_RAWDUMP_HEADER, SAMPLE_PHASED_HEADER])
    def test_pickle_header(self, sample):
        with open(sample, 'rt') as fh:
            header = gsb.GSBHeader.fromfile(fh, verify=True)

        pickled = pickle.dumps(header)
        recovered = pickle.loads(pickled)
        assert recovered == header

    @pytest.mark.parametrize('sample,mode', [
        (SAMPLE_RAWDUMP_HEADER, 'rawdump'),
        (SAMPLE_PHASED_HEADER, 'phased')])
    def test_raw_info(self, sample, mode):
        with gsb.open(sample, 'rt') as fh:
            expected = len(fh.fh_raw.readlines())
            fh.seek(0)
            header0 = gsb.GSBHeader.fromfile(fh, verify=True)
            info = fh.info
            assert info.format == 'gsb'
            assert info.mode == mode
            assert info.number_of_frames == expected
            assert abs(info.frame_rate - u.Hz / 0.251658240) < 1. * u.nHz
            assert info.start_time == header0.time
            assert info.readable is None
            assert info.missing.keys() == {'raw'}
            assert info.errors == {}
            assert info.warnings == {}

    def test_decoding(self):
        """Check that 4-bit encoding works."""
        areal = np.arange(-8., 8.)
        b = encode_4bit(areal)
        assert np.all(b.view(np.uint8) == np.array(
            [0x98, 0xba, 0xdc, 0xfe, 0x10, 0x32, 0x54, 0x76]))
        d = decode_4bit(b)
        assert np.all(d == areal)
        # Also check shape is preserved.
        areal2 = np.hstack((areal, areal[::-1]))
        b2 = encode_4bit(areal2)
        assert np.all(b2.view(np.uint8) == np.array(
            [0x98, 0xba, 0xdc, 0xfe, 0x10, 0x32, 0x54, 0x76,
             0x67, 0x45, 0x23, 0x01, 0xef, 0xcd, 0xab, 0x89]))
        d2 = decode_4bit(b2)
        assert np.all(d2 == areal2)

    def test_payload(self):
        with open(SAMPLE_RAWDUMP, 'rb') as fh:
            # For rawdump, payload_nbytes = (
            #   8192 samples/frame * 0.5 bytes/sample)
            payload1 = gsb.GSBPayload.fromfile(
                fh, payload_nbytes=self.payload_nbytes)
            # Values from reading original raw file.
            assert np.all(payload1.data[:20].ravel() == np.array(
                [0., -2., -2., 0., 4., -1., -2., -1., 1., 2., -1., 1.,
                 -1., 1., -2., 0., -1., -2., 1., -1.], dtype=np.float32))
            assert payload1.data.shape == (8192, 1)
            assert payload1.sample_shape == (1,)
            assert payload1.sample_shape.nchan == 1
            assert payload1.shape == (8192, 1)
            assert payload1.size == 8192
            assert payload1.ndim == 2
            with pytest.raises(ValueError):
                gsb.GSBPayload.fromfile(fh, payload_nbytes=None)
            payload2 = gsb.GSBPayload.fromdata(payload1.data, bps=4)
            assert np.all(payload2.data == payload1.data)
            payload3 = gsb.GSBPayload(payload1.words, bps=4,
                                      sample_shape=payload1.sample_shape)
            assert np.all(payload3.data == payload1.data)

        with open(SAMPLE_PHASED[0][0], 'rb') as fh:
            # For 1 file in phased, payload_nbytes = (
            #   4 full samples/file * 512 channels/full sample *
            #   2 complex elements/channel * 1 bytes/element)
            payload4 = gsb.GSBPayload.fromfile(
                fh, bps=8, complex_data=True,
                payload_nbytes=self.payload_nbytes)
            assert np.all(payload4.data[:20].ravel() == np.array(
                [30. + 12.j, -1. + 8.j, 7. + 19.j, -25. - 5.j, 26. + 14.j,
                 -9. + 0.j, -4. - 1.j, 7. + 6.j, 3. + 5.j, 1. - 2.j,
                 1. - 5.j, 10. - 6.j, 15. - 11.j, -6. + 13.j, 7. + 0.j,
                 -10. - 1.j, -8. + 7.j, 13. + 7.j, -1. + 1.j, 0. + 4.j],
                dtype=np.complex64))
            assert payload4.data.shape == (2048, 1)
            assert payload4.sample_shape == (1,)
            assert payload4.sample_shape.nchan == 1
            payload5 = gsb.GSBPayload.fromdata(payload4.data, bps=8)
            assert np.all(payload5.words == payload4.words)
            payload6 = gsb.GSBPayload(payload4.words, bps=8, complex_data=True,
                                      sample_shape=payload4.sample_shape)
            assert np.all(payload6.data == payload4.data)

        # Test encoding then decoding consistency via `fromdata`.
        payload7 = gsb.payload.GSBPayload.fromdata(payload4.data.real, bps=8)
        assert np.all(payload7.data == payload4.data.real)
        # Encode complex data.
        payload8 = gsb.GSBPayload.fromdata(payload4.data, bps=8)
        assert np.all(payload8.data == payload4.data)
        assert np.all(payload8.words == payload8.words)
        # Encode channelized data.
        channelized = payload4.data.reshape(-1, 512)
        payload9 = gsb.GSBPayload.fromdata(channelized, bps=8)
        assert payload9.shape == channelized.shape
        assert payload9.sample_shape == (512,)
        assert payload9.sample_shape.nchan == 512
        assert np.all(payload9.words == payload4.words)

    def test_phased_payload(self):
        """Test passing tuple into GSBPayload.fromfile.

        (Check that the same result is obtained with a set of single files.)
        """
        # Open tuple of tuple of binary filehandles.
        fh = [[open(thread, 'rb') for thread in pol]
              for pol in SAMPLE_PHASED]
        phased = gsb.GSBPayload.fromfile(
            fh, payload_nbytes=self.payload_nbytes, sample_shape=(2, 512),
            bps=8, complex_data=True)
        assert phased.shape == (8, 2, 512)
        assert phased.sample_shape == (2, 512)
        # Open equivalent sequence of individual payloads, and extract data.
        idata = np.empty([2, 2, 2048], dtype=np.complex64)
        for i, pol in enumerate(SAMPLE_PHASED):
            for j, thread in enumerate(pol):
                with open(thread, 'rb') as ft:
                    ftpayload = gsb.GSBPayload.fromfile(
                        ft, payload_nbytes=self.payload_nbytes,
                        bps=8, complex_data=True)
                    idata[i, j] = ftpayload.data[:, 0]
        # Channelize and merge threads.
        idata = idata.reshape(2, 8, 512).transpose(1, 0, 2)
        assert np.all(phased.data == idata)
        # Raises error for incorrect bps * nchan.
        with pytest.raises(AssertionError):
            gsb.GSBPayload.fromfile(fh, payload_nbytes=self.payload_nbytes,
                                    sample_shape=(12, 1), bps=4)
        # Close file handles.
        for pol in fh:
            for thread in pol:
                thread.close()

    def test_rawdump_frame(self, tmpdir):
        # Load rawdump frame.
        with open(SAMPLE_RAWDUMP_HEADER, 'rt') as ft, \
                open(SAMPLE_RAWDUMP, 'rb') as fraw:
            frame1 = gsb.GSBFrame.fromfile(ft, fraw, bps=4,
                                           payload_nbytes=self.payload_nbytes)
        # Open sections individually.
        with open(SAMPLE_RAWDUMP_HEADER, 'rt') as fh:
            header1 = gsb.GSBHeader.fromfile(fh, verify=True)
        with open(SAMPLE_RAWDUMP, 'rb') as fh:
            payload1 = gsb.GSBPayload.fromfile(
                fh, payload_nbytes=self.payload_nbytes)
        # Compare the two.
        assert header1 == frame1.header
        assert np.all(payload1.data == frame1.payload.data)
        assert frame1.shape == payload1.shape
        assert frame1.size == payload1.size
        assert frame1.ndim == payload1.ndim

        frame2 = gsb.GSBFrame(frame1.header, frame1.payload)
        assert frame2 == frame1
        with open(str(tmpdir.join('test.timestamp')), 'w+t') as sh, \
                open(str(tmpdir.join('test.dat')), 'w+b') as sp:
            frame1.tofile(sh, sp)
            sh.seek(0)
            sp.seek(0)
            frame3 = gsb.GSBFrame.fromfile(sh, sp, bps=4,
                                           payload_nbytes=frame1.nbytes)
        assert frame3 == frame1

    def seek_phased_rawfiles(self, fraw, offset):
        for pol in fraw:
            for thread in pol:
                thread.seek(offset)

    def close_phased_rawfiles(self, fraw):
        for pol in fraw:
            for thread in pol:
                thread.close()

    def test_phased_frame(self, tmpdir):
        # Load phased frame.
        fraw = [[open(thread, 'rb') for thread in pol]
                for pol in SAMPLE_PHASED]
        with open(SAMPLE_PHASED_HEADER, 'rt') as ft:
            frame1 = gsb.GSBFrame.fromfile(ft, fraw,
                                           payload_nbytes=self.payload_nbytes,
                                           sample_shape=(2, 512), bps=8,
                                           complex_data=True)
        self.seek_phased_rawfiles(fraw, 0)
        # Open sections individually.
        with open(SAMPLE_PHASED_HEADER, 'rt') as fh:
            header1 = gsb.GSBHeader.fromfile(fh, verify=True)
        payload1 = gsb.GSBPayload.fromfile(
            fraw, payload_nbytes=self.payload_nbytes,
            sample_shape=(2, 512), bps=8, complex_data=True)
        # Compare the two.
        assert frame1.dtype.kind == 'c'
        assert header1 == frame1.header
        assert frame1.shape == payload1.shape
        assert frame1.size == payload1.size
        assert frame1.ndim == payload1.ndim
        assert np.all(frame1.payload.data == payload1.data)
        assert frame1.valid is True
        frame1.valid = False
        assert frame1.valid is False
        assert np.all(frame1.data == 0.)
        frame1.valid = True
        assert frame1.valid is True
        assert np.all(frame1.payload.data == payload1.data)
        self.close_phased_rawfiles(fraw)

        # Try only right polarization.
        fraw = [[open(thread, 'rb') for thread in SAMPLE_PHASED[1]]]
        with open(SAMPLE_PHASED_HEADER, 'rt') as ft:
            frame2 = gsb.GSBFrame.fromfile(
                ft, fraw, payload_nbytes=self.payload_nbytes,
                sample_shape=(1, 512), bps=8, complex_data=True)
        self.seek_phased_rawfiles(fraw, 0)
        payload2 = gsb.GSBPayload.fromfile(
            fraw, payload_nbytes=self.payload_nbytes,
            sample_shape=(1, 512), bps=8, complex_data=True)
        assert frame2.shape == payload2.shape
        assert np.all(frame2.payload.data == payload2.data)
        self.close_phased_rawfiles(fraw)

        frame3a = gsb.GSBFrame.fromdata(payload1.data, header1, bps=8)
        assert frame3a.shape == frame1.shape
        assert np.all(frame3a.data == frame1.data)
        frame3b = gsb.GSBFrame.fromdata(payload1.data, bps=8, **header1)
        assert frame3b.shape == frame1.shape
        assert np.all(frame3b.data == frame1.data)
        # Test writing phased data to multiple files.
        with open(str(tmpdir.join('test.timestamp')), 'w+t') as sh, \
                open(str(tmpdir.join('test0.dat')), 'w+b') as sp0, \
                open(str(tmpdir.join('test1.dat')), 'w+b') as sp1, \
                open(str(tmpdir.join('test2.dat')), 'w+b') as sp2, \
                open(str(tmpdir.join('test3.dat')), 'w+b') as sp3:
            frame1.tofile(sh, ((sp0, sp1), (sp2, sp3)))
            for s in sp0, sp1, sp2, sp3:
                s.flush()
                s.seek(0)
            sh.flush()
            sh.seek(0)
            frame4 = gsb.GSBFrame.fromfile(
                sh, ((sp0, sp1), (sp2, sp3)),
                payload_nbytes=self.payload_nbytes,
                sample_shape=(2, 512), bps=8, complex_data=True)
        assert frame4 == frame1

    @pytest.mark.parametrize('sample', (SAMPLE_RAWDUMP_HEADER,
                                        SAMPLE_PHASED_HEADER))
    def test_timestamp_io(self, tmpdir, sample):
        """Tests GSBTimeStampIO in base.py."""
        with open(sample, 'rt') as fh:
            header0 = gsb.GSBHeader.fromfile(fh, verify=True)

        with gsb.open(sample, 'rt') as fh:
            header1 = fh.read_timestamp()
            assert header1 == header0
            current_pos = fh.tell()
            frame_rate = fh.get_frame_rate()
            assert abs(frame_rate - u.Hz / 0.251658240) < 1. * u.nHz
            assert fh.tell() == current_pos

        testfile = str(tmpdir.join('test.timestamp'))
        with gsb.open(testfile, 'wt') as fw:
            fw.write_timestamp(header=header1)
            fw.write_timestamp(mode=header1.mode, **header1)
        with gsb.open(testfile, 'rt') as fh:
            header2 = fh.read_timestamp()
            assert header2 == header1
            header3 = fh.read_timestamp()
            assert header3 == header1

        # Check that extra arguments raise TypeError.
        with pytest.raises(TypeError):
            gsb.open(testfile, 'rt', raw='bla')

    @pytest.mark.parametrize('sample', (SAMPLE_RAWDUMP_HEADER,
                                        SAMPLE_PHASED_HEADER))
    def test_pickle_timestamp_io(self, sample):
        """Tests GSBTimeStampIO in base.py."""
        with gsb.open(sample, 'rt') as fh:
            fh.read_timestamp()
            pickled = pickle.dumps(fh)
            header1 = fh.read_timestamp()

        with pickle.loads(pickled) as fh2:
            header2 = fh2.read_timestamp()

        assert header2 == header1

    def test_rawfile_io(self, tmpdir):
        """Tests GSBFileReader and GSBFileWriter in base.py."""
        with open(SAMPLE_RAWDUMP, 'rb') as fh:
            payload1 = gsb.GSBPayload.fromfile(
                fh, payload_nbytes=self.payload_nbytes)
        testfile = str(tmpdir.join('test.dat'))
        with gsb.open(testfile, 'wb') as fw:
            assert fw.writable()
            fw.write_payload(payload1, bps=4)
            fw.write_payload(payload1.data, bps=4)
        with gsb.open(testfile, 'rb', payload_nbytes=2**12) as fh:
            assert fh.readable()
            assert not fh.writable()
            payload2 = fh.read_payload()
            assert payload2 == payload1
            payload3 = fh.read_payload()
            assert payload3 == payload1

    def test_pickle_filereader(self):
        """Tests GSBFileReader in base.py."""
        with gsb.open(SAMPLE_RAWDUMP, 'rb', payload_nbytes=2**12, nchan=1,
                      bps=4, complex_data=False) as fh:
            fh.read_payload()
            pickled = pickle.dumps(fh)
            payload1 = fh.read_payload()

        with pickle.loads(pickled) as fh2:
            payload2 = fh2.read_payload()

        assert payload2 == payload1

    def test_rawfile_repr(self):
        with gsb.open(SAMPLE_RAWDUMP, 'rb', payload_nbytes=2**12, nchan=1,
                      bps=4, complex_data=False) as fh:
            repr_fh = repr(fh)
        assert repr_fh.startswith('GSBFileReader')
        assert ('payload_nbytes=4096, nchan=1, bps=4, complex_data=False' in
                repr_fh)

    def test_raw_stream(self, tmpdir):
        bps = 4
        sample_rate = self.frame_rate * self.payload_nbytes * (8 // bps)
        # Open here with payloadsize given, below with samples_per_frame.
        with gsb.open(SAMPLE_RAWDUMP_HEADER, 'rs', raw=SAMPLE_RAWDUMP,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes,
                      squeeze=False) as fh_r:
            assert fh_r.readable()
            # Get some attributes from fh_raw but not all.
            assert fh_r.seekable()
            assert not fh_r.writable()
            assert not hasattr(fh_r, 'read_payload')
            # Also checks repr works at all...
            assert 'rawdump' in repr(fh_r)

            with open(SAMPLE_RAWDUMP_HEADER, 'rt') as ft, \
                    open(SAMPLE_RAWDUMP, 'rb') as fraw:
                frame1 = gsb.GSBFrame.fromfile(
                    ft, fraw, bps=4, payload_nbytes=self.payload_nbytes)
            assert fh_r.header0.time == fh_r.start_time
            assert fh_r.header0 == frame1.header
            assert fh_r.sample_shape == (1,)
            assert fh_r.shape == ((10 * fh_r.samples_per_frame,)
                                  + fh_r.sample_shape)
            assert fh_r.size == np.prod(fh_r.shape)
            assert fh_r.ndim == len(fh_r.shape)
            assert fh_r.sample_rate == sample_rate
            check = fh_r.read(fh_r.samples_per_frame)
            assert np.all(check == frame1.data)
            # Seek last offset.
            with open(SAMPLE_RAWDUMP_HEADER, 'rt') as ft, \
                    open(SAMPLE_RAWDUMP, 'rb') as fraw:
                ft.seek(frame1.header.seek_offset(9))
                fraw.seek(9 * fh_r.payload_nbytes)
                frame10 = gsb.GSBFrame.fromfile(
                    ft, fraw, bps=4, payload_nbytes=self.payload_nbytes)
            assert fh_r._last_header == frame10.header
            fh_r.seek(-10, 2)
            check = fh_r.read(10)
            assert np.all(check == frame10.data[-10:])
            # Check validity of current and stopping time
            assert abs(fh_r.stop_time
                       - Time('2015-04-27T13:15:02.516582640')) < 1.*u.ns
            assert abs(fh_r.stop_time - fh_r.time) < 1.*u.ns
            fh_r.seek(0)
            data1 = fh_r.read()
            assert fh_r.tell() == len(data1)
            assert data1.shape == fh_r.shape
            fh_r.seek(0)
            out1 = np.empty_like(data1)
            fh_r.read(out=out1)
            assert np.all(out1 == data1)

        # Try with squeezing on.
        with gsb.open(SAMPLE_RAWDUMP_HEADER, 'rs', raw=SAMPLE_RAWDUMP,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes,
                      squeeze=True) as fh_r:
            data2 = fh_r.read()
            assert np.all(data2 == data1.squeeze())
            out2 = np.empty(fh_r.shape)
            fh_r.seek(0)
            fh_r.read(out=out2)
            assert np.all(out2 == data1.squeeze())
            # To compare with directly psasing samples_per_frame below.
            spf_from_payload_nbytes = fh_r.samples_per_frame
            # For testing proper file closing below.
            header0 = fh_r.header0

        gsbtest_ts = str(tmpdir.join('test_time.timestamp'))
        gsbtest_raw = str(tmpdir.join('test.dat'))
        with gsb.open(SAMPLE_RAWDUMP_HEADER, 'rs', raw=SAMPLE_RAWDUMP,
                      sample_rate=sample_rate, samples_per_frame=(
                          self.payload_nbytes * (8 // bps))) as fh_r:
            # Check that passing samples_per_frame is identical to passing
            # payload_nbytes.
            assert fh_r.samples_per_frame == spf_from_payload_nbytes
            check = fh_r.read()
            assert np.all(check == data2)

            with gsb.open(gsbtest_ts, 'ws', raw=gsbtest_raw,
                          header0=fh_r.header0, sample_rate=fh_r.sample_rate,
                          samples_per_frame=(self.payload_nbytes * (8 // bps))
                          ) as fh_w:
                assert fh_w.sample_rate == sample_rate
                fh_w.write(data2)

            with gsb.open(gsbtest_ts, 'rs', raw=gsbtest_raw,
                          sample_rate=fh_r.sample_rate,
                          samples_per_frame=(
                              self.payload_nbytes * (8 // bps))) as fh_n:
                assert fh_n.header0 == fh_r.header0
                assert fh_n._last_header == fh_r._last_header
                assert fh_n.sample_shape == fh_r.sample_shape
                assert fh_n.shape == fh_r.shape
                assert fh_n.start_time == fh_r.start_time
                assert fh_n.sample_rate == sample_rate
                check = fh_n.read()
                assert np.all(check == data2)
                assert abs(fh_n.stop_time - fh_n.time) < 1.*u.ns
                assert abs(fh_n.stop_time - fh_r.stop_time) < 1.*u.ns

        # Try writing with squeeze off, and reading it back with squeeze.
        with gsb.open(gsbtest_ts, 'ws', raw=gsbtest_raw,
                      header0=header0, sample_rate=sample_rate,
                      samples_per_frame=(self.payload_nbytes * (8 // bps)),
                      squeeze=False) as fh_wns:
            fh_wns.write(data1)

        with gsb.open(gsbtest_ts, 'rs', raw=gsbtest_raw,
                      sample_rate=sample_rate,
                      samples_per_frame=(self.payload_nbytes
                                         * (8 // bps))) as fh_nns:
            check == fh_nns.read()
            assert np.all(check == data2)

        # Test that opening that raises an exception correctly handles
        # file closing.
        with open(str(tmpdir.join('test.timestamp')), 'wt') as sh, \
                open(str(tmpdir.join('test.dat')), 'w+b') as sp:
            with pytest.raises(u.UnitsError):
                gsb.open(sh, 'ws', raw=sp, sample_rate=3.9736/u.m,
                         samples_per_frame=(self.payload_nbytes * (8 // bps)),
                         **header0)
            assert not sh.closed
            assert not sp.closed

        # Test not passing a sample rate and samples per frame to reader
        # (can't test reading, since the sample file is tiny).
        with gsb.open(SAMPLE_RAWDUMP_HEADER, 'rs', raw=SAMPLE_RAWDUMP) as fh_r:
            assert u.isclose(fh_r.sample_rate, (100. / 3.) * u.MHz,
                             rtol=2**-52)
            assert fh_r.samples_per_frame == 2**23
            assert fh_r.payload_nbytes == 2**22

        # Test that missing header line will lead to failure.
        filename_missinghead = str(tmpdir.join('missing.timestamp'))
        with open(SAMPLE_RAWDUMP_HEADER, 'rt') as fh_r:
            timestamps = fh_r.readlines()
            del timestamps[3]
        with open(filename_missinghead, 'wt') as fh_w:
            fh_w.writelines(timestamps)
        with gsb.open(filename_missinghead, 'rs', raw=SAMPLE_RAWDUMP,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes) as fr:
            with pytest.raises(ValueError, match='wrong frame'):
                fr.read()

    @pytest.mark.parametrize('bad', [False, True])
    def test_bad_last_timestamp(self, bad, tmpdir):
        # Test that an incomplete or bad last header leads to the
        # second-to-last header being used, and raises a warning.
        filename_incompletehead = str(
            tmpdir.join('test_incomplete_header.timestamp'))
        with open(SAMPLE_RAWDUMP_HEADER, 'rt') as fh, \
                open(filename_incompletehead, 'wt') as fw:
            fw.write(fh.read()[:-4])
            if bad:
                fw.write('xxxx')

        with gsb.open(filename_incompletehead, 'rt') as fh_t:
            assert 'number_of_frames' in fh_t.info.warnings
            warn_exp = 'failed to read' if bad else 'incomplete'
            assert warn_exp in fh_t.info.warnings['number_of_frames']

        with gsb.open(filename_incompletehead, 'rs', raw=SAMPLE_RAWDUMP,
                      payload_nbytes=self.payload_nbytes,
                      squeeze=False) as fh_r:
            with pytest.warns(UserWarning, match='second-to-last entry'):
                fh_r._last_header
            assert fh_r.shape[0] == 9 * fh_r.samples_per_frame
            assert warn_exp in fh_r.info.warnings['number_of_frames']

        with open(SAMPLE_RAWDUMP_HEADER, 'rt') as fh, \
                open(filename_incompletehead, 'wt') as fw:
            fw.write(fh.read()[:45])
        with gsb.open(filename_incompletehead, 'rs', raw=SAMPLE_RAWDUMP,
                      payload_nbytes=self.payload_nbytes,
                      squeeze=False) as fh_r:
            with pytest.warns(UserWarning, match='second-to-last entry'):
                fh_r._last_header
            assert fh_r.shape[0] == fh_r.samples_per_frame
            assert fh_r._last_header == fh_r.header0

        # Test not passing a sample rate and samples per frame to reader
        # (can't test reading, since the sample file is tiny).
        with gsb.open(SAMPLE_RAWDUMP_HEADER, 'rs', raw=SAMPLE_RAWDUMP) as fh_r:
            assert u.isclose(fh_r.sample_rate, (100. / 3.) * u.MHz,
                             rtol=2**-52)
            assert fh_r.samples_per_frame == 2**23
            assert fh_r.payload_nbytes == 2**22

    @pytest.mark.parametrize('sample_header,sample_data', [
        (SAMPLE_RAWDUMP_HEADER, SAMPLE_RAWDUMP),
        (SAMPLE_PHASED_HEADER, SAMPLE_PHASED)])
    def test_pickle(self, sample_header, sample_data):
        if sample_header is SAMPLE_RAWDUMP_HEADER:
            sample_rate = self.frame_rate * self.payload_nbytes * 2
        else:
            sample_rate = self.frame_rate * self.payload_nbytes / 512
        # Only simple tests here; more complete ones in vdif.
        with gsb.open(sample_header, 'rs', raw=sample_data,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes,
                      squeeze=False) as fh:
            fh.seek(6)
            pickled = pickle.dumps(fh)
            d1_3 = fh.read(3)
            with pickle.loads(pickled) as fh2:
                assert fh2.tell() == 6
                d2_10 = fh2.read(10)

            assert np.all(d2_10[:3] == d1_3)
            assert fh.tell() == 9

        with pickle.loads(pickled) as fh3:
            assert fh3.tell() == 6
            d3_5 = fh3.read(5)

        assert np.all(d3_5[:3] == d1_3)

        closed = pickle.dumps(fh)
        with pickle.loads(closed) as fh4:
            assert fh4.closed
            with pytest.raises(ValueError):
                fh4.read(1)

    @pytest.mark.parametrize('sample_header,sample_data', [
        (SAMPLE_RAWDUMP_HEADER, SAMPLE_RAWDUMP),
        (SAMPLE_PHASED_HEADER, SAMPLE_PHASED)])
    def test_copy(self, sample_header, sample_data):
        if sample_header is SAMPLE_RAWDUMP_HEADER:
            sample_rate = self.frame_rate * self.payload_nbytes * 2
        else:
            sample_rate = self.frame_rate * self.payload_nbytes / 512
        # Only simple tests here; more complete ones in vdif.
        with gsb.open(sample_header, 'rs', raw=sample_data,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes,
                      squeeze=False) as fh:
            fh.seek(6)
            with copy.deepcopy(fh) as fh2:
                d1_3 = fh.read(3)
                assert fh2.tell() == 6
                d2_10 = fh2.read(10)
                assert fh.tell() == 9
                assert fh2.tell() == 16

            assert np.all(d2_10[:3] == d1_3)
            assert fh2.closed
            assert not fh.closed
            d1_7 = fh.read(7)
            assert np.all(d2_10[3:] == d1_7)

        with copy.copy(fh) as fh3:
            assert fh3.closed

    def test_phased_stream(self, tmpdir):
        bps = 8
        nchan = 512
        sample_shape = (2, 512)
        sample_rate = (self.frame_rate * self.payload_nbytes
                       * (8 // bps) / nchan)
        # Open here with payloadsize given, below with samples_per_frame.
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=SAMPLE_PHASED,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes,
                      squeeze=False) as fh_r:
            assert fh_r.readable()
            # Get some attributes from fh_raw but not all.
            assert fh_r.seekable()
            assert not fh_r.writable()
            assert not hasattr(fh_r, 'read_payload')
            # Also checks repr works at all...
            assert 'phased' in repr(fh_r)

            fraw = [[open(thread, 'rb') for thread in pol]
                    for pol in SAMPLE_PHASED]
            with open(SAMPLE_PHASED_HEADER, 'rt') as ft:
                frame1 = gsb.GSBFrame.fromfile(
                    ft, fraw, payload_nbytes=self.payload_nbytes,
                    sample_shape=sample_shape, bps=bps, complex_data=True)
            assert fh_r.header0.time == fh_r.start_time
            assert fh_r.header0 == frame1.header
            assert fh_r.sample_shape == sample_shape
            assert fh_r.shape == ((10 * fh_r.samples_per_frame,)
                                  + fh_r.sample_shape)
            assert fh_r.size == np.prod(fh_r.shape)
            assert fh_r.ndim == len(fh_r.shape)
            assert fh_r.sample_rate == sample_rate
            assert np.all(fh_r.read(fh_r.samples_per_frame) == frame1.data)
            # Seek last offset.
            with open(SAMPLE_PHASED_HEADER, 'rt') as ft:
                ft.seek(frame1.header.seek_offset(9))
                self.seek_phased_rawfiles(fraw, 9 * fh_r.payload_nbytes)
                frame10 = gsb.GSBFrame.fromfile(
                    ft, fraw, payload_nbytes=self.payload_nbytes,
                    sample_shape=sample_shape, bps=bps, complex_data=True)
            assert fh_r._last_header == frame10.header
            fh_r.seek(-8, 2)
            assert np.all(fh_r.read(8) == frame10.data)
            # Check validity of current and stopping time.
            assert abs(fh_r.stop_time
                       - Time('2013-07-27T21:23:57.8406912')) < 1.*u.ns
            assert abs(fh_r.stop_time - fh_r.time) < 1.*u.ns
            fh_r.seek(0)
            data1 = fh_r.read()
            assert fh_r.tell() == len(data1)
            assert data1.shape == fh_r.shape
            fh_r.seek(0)
            out1 = np.empty_like(data1)
            fh_r.read(out=out1)
            assert np.all(out1 == data1)
            fh_r.seek(1, 'end')
            with pytest.raises(EOFError):
                fh_r.read()

        # Try again with squeezing.
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=SAMPLE_PHASED,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes,
                      squeeze=True) as fh_r:
            out2 = np.empty(fh_r.shape, dtype=np.complex64)
            fh_r.read(out=out2)
            assert np.all(out2 == out1.squeeze())
            # To compare with directly psasing samples_per_frame below.
            spf_from_payload_nbytes = fh_r.samples_per_frame
            self.close_phased_rawfiles(fraw)

        # Try only right polarization.
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=SAMPLE_PHASED[1],
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes,
                      squeeze=False) as fh_r:

            fraw = [[open(thread, 'rb') for thread in SAMPLE_PHASED[1]]]
            with open(SAMPLE_PHASED_HEADER, 'rt') as ft:
                frame1 = gsb.GSBFrame.fromfile(
                    ft, fraw, payload_nbytes=self.payload_nbytes,
                    sample_shape=(1, nchan), bps=bps, complex_data=True)
            assert fh_r.header0.time == fh_r.start_time
            assert fh_r.header0 == frame1.header
            assert np.all(fh_r.read(fh_r.samples_per_frame) == frame1.data)
            self.close_phased_rawfiles(fraw)

        # Test subsetting.
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=SAMPLE_PHASED,
                      sample_rate=sample_rate, subset=(1, 3),
                      payload_nbytes=self.payload_nbytes) as fh_r:
            assert fh_r.sample_shape == ()
            assert np.all(fh_r.read() == data1[:, 1, 3])

        subset_md = (np.array([1, 0])[:, np.newaxis], [1, 33, 121, 245])
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=SAMPLE_PHASED,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes,
                      subset=subset_md) as fh_r:
            assert fh_r.sample_shape == (2, 4)
            assert np.all(fh_r.read() == data1[(slice(None),) + subset_md])

        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=SAMPLE_PHASED[1],
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes,
                      subset=slice(0, 256)) as fh_r:
            assert fh_r.sample_shape == (256,)
            fraw = [[open(thread, 'rb') for thread in SAMPLE_PHASED[1]]]
            with open(SAMPLE_PHASED_HEADER, 'rt') as ft:
                frame1 = gsb.GSBFrame.fromfile(
                    ft, fraw, payload_nbytes=self.payload_nbytes,
                    sample_shape=(1, nchan),
                    bps=bps, complex_data=True)
            assert np.all(fh_r.read(fh_r.samples_per_frame)
                          == frame1.data[:, 0, :256])
            self.close_phased_rawfiles(fraw)

        # Try writing to file by passing header keywords into open.
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=SAMPLE_PHASED,
                      sample_rate=sample_rate,
                      samples_per_frame=(self.payload_nbytes
                                         // nchan)) as fh_r, \
                open(str(tmpdir.join('test_time.timestamp')), 'w+t') as sh,\
                open(str(tmpdir.join('test0.dat')), 'w+b') as sp0, \
                open(str(tmpdir.join('test1.dat')), 'w+b') as sp1, \
                open(str(tmpdir.join('test2.dat')), 'w+b') as sp2, \
                open(str(tmpdir.join('test3.dat')), 'w+b') as sp3:
            fraw = ((sp0, sp1), (sp2, sp3))
            fh_w = gsb.open(sh, 'ws', raw=fraw, sample_rate=fh_r.sample_rate,
                            samples_per_frame=(self.payload_nbytes // nchan),
                            **fh_r.header0)
            assert fh_w.sample_rate == sample_rate
            fh_w.write(fh_r.read())
            fh_w.flush()
            sh.seek(0)
            self.seek_phased_rawfiles(fraw, 0)
            fh_r.seek(0)
            with gsb.open(sh, 'rs', raw=fraw, sample_rate=sample_rate,
                          samples_per_frame=(self.payload_nbytes // nchan)
                          ) as fh_n:
                assert fh_n.header0 == fh_r.header0
                # PC time will not be the same.
                for key in ('gps', 'seq_nr', 'mem_block'):
                    assert fh_n._last_header[key] == fh_r._last_header[key]
                assert fh_n.shape == fh_r.shape
                assert fh_n.sample_shape == fh_r.sample_shape
                assert fh_n.start_time == fh_r.start_time
                assert fh_n.sample_rate == sample_rate
                assert np.all(fh_n.read() == fh_r.read())
                assert abs(fh_n.stop_time - fh_n.time) < 1.*u.ns
                assert abs(fh_n.stop_time - fh_r.stop_time) < 1.*u.ns
            # Check that passing samples_per_frame is identical to passing
            # payload_nbytes.
            fh_r.seek(0)
            assert fh_r.samples_per_frame == spf_from_payload_nbytes
            assert np.all(fh_r.read() == data1)
            fh_w.close()

    @pytest.mark.parametrize('raw', [SAMPLE_PHASED, SAMPLE_PHASED[:1]])
    def test_phased_stream_one_file_per_pol(self, raw):
        sample_rate = self.frame_rate * self.payload_nbytes / 512
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=raw,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes) as fh_2file:
            full_data = fh_2file.read()

        raw_one_file = [pol_files[:1] for pol_files in raw]
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=raw_one_file,
                      sample_rate=sample_rate/2,
                      payload_nbytes=self.payload_nbytes) as fh_1file:
            data = fh_1file.read()
        assert data.shape[0] == full_data.shape[0] // 2
        assert data.shape[1:] == full_data.shape[1:]
        samples_per_block = fh_2file.samples_per_frame // 2
        assert samples_per_block == fh_1file.samples_per_frame
        blocked = full_data.reshape((-1, 2, samples_per_block)
                                    + fh_1file.sample_shape)
        expected = blocked[:, 0].reshape((-1,)+fh_1file.sample_shape)
        assert np.all(data == expected)

    def test_phased_stream_one_file(self):
        raw = [[SAMPLE_PHASED[0][0]]]
        sample_rate = self.frame_rate * self.payload_nbytes / 512 / 2
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=raw,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes) as fh:
            ref_data = fh.read()

        raw_one_file = raw[0][0]
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=raw_one_file,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes) as fh_1file:
            assert fh_1file.header0.mode == 'phased'
            data = fh_1file.read()

        assert np.all(data == ref_data)

    def test_phased_write_one_file(self, tmpdir):
        # With a single file, one does have to pass in header_mode
        # to get phased.
        with gsb.open(str(tmpdir.join('test.timstamp')), 'ws',
                      raw=str(tmpdir.join('test.raw')),
                      header_mode='phased',
                      time=Time('2010-10-10')) as fh_right:
            assert fh_right.header0.mode == 'phased'

        with gsb.open(str(tmpdir.join('test.timstamp')), 'ws',
                      raw=str(tmpdir.join('test.raw')),
                      time=Time('2010-10-10')) as fh_wrong:
            assert fh_wrong.header0.mode == 'rawdump'

    @pytest.mark.parametrize('stop, nframes', [(-7, 9), (97, 1)])
    def test_stream_incomplete_header(self, stop, nframes, tmpdir):
        # Test that an incomplete last header leads to the second-to-last
        # header being used, and raises a warning.
        sample_rate = self.frame_rate * self.payload_nbytes / 512
        filename_incompletehead = str(
            tmpdir.join('test_incomplete_header.timestamp'))
        with open(SAMPLE_PHASED_HEADER, 'rt') as fh, \
                open(filename_incompletehead, 'wt') as fw:
            fw.write(fh.read()[:stop])
        with gsb.open(filename_incompletehead, 'rs', raw=SAMPLE_PHASED,
                      sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes,
                      squeeze=False) as fh_r:
            with pytest.warns(UserWarning, match='second-to-last entry'):
                shape = fh_r.shape

            assert shape[0] == nframes * fh_r.samples_per_frame

            info = fh_r.info
            assert info.errors == {}
            assert info.warnings.keys() == {'number_of_frames',
                                            'consistent'}
            assert 'incomplete' in info.warnings['number_of_frames']
            assert 'contains more bytes' in info.warnings['consistent']

    @pytest.mark.parametrize('raw, nstream', [
        (SAMPLE_PHASED, 2),
        (SAMPLE_PHASED[:1], 2),
        ((SAMPLE_PHASED[0][:1], SAMPLE_PHASED[1][:1]), 1),
        (SAMPLE_PHASED[1][1], 1),
    ])
    def test_stream_reader_defaults(self, raw, nstream):
        # Test not passing a sample rate and samples per frame to reader
        # (can't test reading, since the sample file is tiny).
        default_frame_rate = (100/6/2**22)*u.MHz
        with gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=raw) as fh_r:
            assert fh_r.sample_shape[-1] == 512
            assert fh_r.payload_nbytes == 2**22
            assert fh_r.samples_per_frame == nstream * 2**12  # 2**22 / 1024
            assert u.isclose(fh_r.sample_rate,
                             fh_r.samples_per_frame*default_frame_rate,
                             rtol=2**-52)

    def test_stream_invalid(self, tmpdir):
        with pytest.raises(Exception):
            gsb.open(SAMPLE_RAWDUMP_HEADER, 'rs', raw=SAMPLE_PHASED,
                     payload_nbytes=self.payload_nbytes)
        with pytest.raises(ValueError):
            # no r or w in mode
            gsb.open('ts.dat', 's')
        with pytest.raises(OSError):
            # non-existing file
            gsb.open(str(tmpdir.join('ts.bla')),
                     raw=str(tmpdir.join('raw.bla')))
        with pytest.raises(TypeError, match="required argument 'raw'"):
            gsb.open(SAMPLE_PHASED_HEADER, 'rs')
        with pytest.raises(ValueError, match='inconsistent'):
            gsb.open(SAMPLE_PHASED_HEADER, 'rs', raw=SAMPLE_PHASED,
                     payload_nbytes=32, samples_per_frame=400)
        with pytest.raises(ValueError, match='inconsistent'):
            gsb.open(SAMPLE_RAWDUMP_HEADER, 'rs', raw=SAMPLE_RAWDUMP,
                     payload_nbytes=32, samples_per_frame=400)

    def test_stream_wrong_payload_warning(self):
        with gsb.open(SAMPLE_RAWDUMP_HEADER, 'rs', raw=SAMPLE_RAWDUMP,
                      payload_nbytes=self.payload_nbytes-1) as fh:
            assert 'consistent' in fh.info.warnings
            assert 'non-integer' in fh.info.warnings['consistent']

    @pytest.mark.parametrize('ts,raw,mode', [
        (SAMPLE_RAWDUMP_HEADER, SAMPLE_RAWDUMP, 'rawdump'),
        (SAMPLE_PHASED_HEADER, SAMPLE_PHASED, 'phased-2pol'),
        (SAMPLE_PHASED_HEADER, SAMPLE_PHASED[0], 'phased-1pol'),
        (SAMPLE_PHASED_HEADER, SAMPLE_PHASED[:1], 'phased-1pol'),
        (SAMPLE_PHASED_HEADER, [SAMPLE_PHASED[0][:1],
                                SAMPLE_PHASED[1][:1]], 'unsplit-2pol'),
        (SAMPLE_PHASED_HEADER, [SAMPLE_PHASED[0][1:]], 'unsplit-1pol'),
        (SAMPLE_PHASED_HEADER, SAMPLE_PHASED[0][1], 'unsplit-1pol'),
    ])
    def test_stream_info(self, ts, raw, mode):
        bps = 4 if mode == 'rawdump' else 8
        nchan = 1 if mode == 'rawdump' else 512
        sample_rate = (self.frame_rate * self.payload_nbytes
                       * (8 // bps) / nchan)
        if mode.startswith('unsplit'):
            sample_rate /= 2
        with gsb.open(ts, 'rs', raw=raw, sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes) as fh:
            info = fh.info
            assert info.format == 'gsb'
            assert info.consistent
            assert info.readable
            assert info.errors == {}
            assert info.warnings == {}
            assert info.file_info.missing == {}
            assert info.checks == {'decodable': True,
                                   'consistent': True}
            assert info.bps == bps
            assert info.payload_nbytes == self.payload_nbytes
            assert u.isclose(info.sample_rate, sample_rate)
            if mode == 'rawdump':
                assert not info.complex_data
                assert info.shape == (81920,)
                assert info.n_raw == 1
            else:
                assert info.complex_data
                if mode.startswith('unsplit'):
                    assert info.shape[0] == 40
                    assert info.n_raw == 1
                else:
                    assert info.shape[0] == 80
                    assert info.n_raw == 2
                if mode.endswith('2pol'):
                    assert info.shape[1:] == (2, nchan)
                else:
                    assert info.shape[1:] == (nchan,)

    @pytest.mark.parametrize('ts,raw', [
        (SAMPLE_PHASED_HEADER, [SAMPLE_PHASED[0][:1],
                                SAMPLE_PHASED[1][:1]]),
        (SAMPLE_PHASED_HEADER, [SAMPLE_PHASED[0][1:]]),
        (SAMPLE_PHASED_HEADER, SAMPLE_PHASED[0][1]),
    ])
    def test_stream_info_inconsistent(self, ts, raw):
        bps = 8
        nchan = 512
        sample_rate = (self.frame_rate * self.payload_nbytes
                       * (8 // bps) / nchan)
        with gsb.open(ts, 'rs', raw=raw, sample_rate=sample_rate,
                      payload_nbytes=self.payload_nbytes, nchan=512) as fh:
            info = fh.info
            assert not fh.info.consistent
            assert isinstance(fh.info.errors['consistent'], EOFError)
            assert 'factor of two' in str(fh.info.errors['consistent'])
            assert info.sample_rate == sample_rate



# Licensed under the GPLv3 - see LICENSE.rst
"""Sample files with baseband data recorded in different formats."""

# Use private names to avoid inclusion in the sphinx documentation.
from os import path as _path


def _full_path(name, dirname=_path.dirname(_path.abspath(__file__))):
    return _path.join(dirname, name)


SAMPLE_MARK4 = _full_path('sample.m4')
"""Mark 4 sample.  ntrack=64, fanout=4, bps=2.

Created from a European VLBI Network/Arecibo PSR B1957+20 observation using
dd if=gp052d_ar_no0021 of=sample.m4 bs=128000 count=3

These data follow standard assignments:
fan_out       0011223300112233001122330011223300112233001122330011223300112233
magnitude_bit 0000000011111111000000001111111100000000111111110000000011111111
lsb_output    1111111111111111111111111111111111111111111111111111111111111111
converter_id  0202020202020202131313131313131346464646464646465757575757575757
"""

SAMPLE_MARK4_32TRACK = _full_path('sample_32track.m4')
"""Mark 4 sample.  ntrack=32, fanout=4, bps=2.

Created from a Arecibo observation simultaneous with RadioAstron using
dd if=rg10a_ar_no0014 of=sample_32track.m4 bs=10000 count=17

These data follow standard assignments:
fan_out       00112233001122330011223300112233
magnitude_bit 00000000111111110000000011111111
lsb_output    01010101010101010101010101010101
converted_id  00000000000000001111111111111111
"""

SAMPLE_MARK4_32TRACK_FANOUT2 = _full_path('sample_32track_fanout2.m4')
"""Mark 4 sample.  ntrack=32, fanout=2, bps=2.

Created from an Arecibo observation of PSR B1133+16 using
dd if=gk049c_ar_no0011.m5a of=sample_32track_fanout2.m4 bs=10000 count=18

These data follow standard assignments:
fan_out       00110011001100110011001100110011
magnitude_bit 00001111000011110000111100001111
lsb_output    00000000000000001111111111111111
converted_id  02020202131313130202020213131313
"""

SAMPLE_MARK4_16TRACK = _full_path('sample_16track.m4')
"""Mark 4 sample.  ntrack=16, fanout=4, bps=2.

Created from the first two frames an Arecibo observation of the Crab Pulsar on
2013/11/03.  (2013_306_raks02ae/ar/gs033a_ar_no0055.m5a)

These data follow standard assignments:
fan_out       0123012301230123
magnitude_bit 0000111100001111
lsb_output    1111111111111111
converted_id  0000000011111111
"""

SAMPLE_MARK4_64TRACK_FANOUT2_FT = _full_path('sample_64track_fanout2_ft.m4')
"""Mark 4 sample. ntrack=64, fanout=2, bps=2.

Created from Fortleza (Ft) station by Shaogang Gao (gh-332).

These data do not follow standard assignments, so need special treatment:
fan_out       0011001100110011001100110011001100110011001100110011001100110011
magnitude_bit 0000010110101111000011110000111100000101101011110000111100001111
lsb_output    0000101000001010000000000000000000001010000010100000000000000000
converted_id  030303030404040415151515262626267a7a7a7a7b7b7b7b8c8c8c8c9d9d9d9d
"""

SAMPLE_MARK5B = _full_path('sample.m5b')
"""Mark 5B sample.  nchan=8, bps=2.

Created from a EVN/WSRT PSR B1957+20 observation.
"""

SAMPLE_VDIF = _full_path('sample.vdif')
"""VDIF sample. 8 threads, bps=2.

Created from a EVN/VLBA PSR B1957+20 observation.  Timestamps of frames with
even thread IDs have been corrected to be consistent with odd-ID frames.
"""

SAMPLE_VLBI_VDIF = _full_path('sample_vlbi.vdif')
"""VDIF sample. 8 threads, bps=2.

Created from a EVN/VLBA PSR B1957+20 observation.  Uncorrected version of
SAMPLE_VDIF.
"""

SAMPLE_MWA_VDIF = _full_path('sample_mwa.vdif')
"""VDIF sample from MWA.  EDV=0, two threads, bps=8"""

SAMPLE_AROCHIME_VDIF = _full_path('sample_arochime.vdif')
"""VDIF sample from ARO, written by CHIME backend. EDV=1, nchan=1024, bps=4."""

SAMPLE_BPS1_VDIF = _full_path('sample_bps1.vdif')
"""VDIF sample from Christian Ploetz. EDV=0, nchan=16, bps=1."""

SAMPLE_DADA = _full_path('sample.dada')
"""DADA sample from Effelsberg, with header adapted to shortened size."""

SAMPLE_PUPPI = _full_path('sample_puppi.raw')
"""GUPPI/PUPPI sample, npol=2, nchan=4.

Created from the first four frames of a 2018-01-14 Arecibo observation of
J1810+1744, with payload shortened to 8192 samples (with 512 overlap),
equivalent to 1024 complete samples (with 64 overlap).
"""

SAMPLE_GSB_RAWDUMP_HEADER = _full_path('gsb/sample_gsb_rawdump.timestamp')
"""GSB rawdump header sample.

First 10 header entries of node 5 rawdump data from 2015-04-27 GMRT
observations of the Crab pulsar.
"""

SAMPLE_GSB_RAWDUMP = _full_path('gsb/sample_gsb_rawdump.dat')
"""GSB rawdump sample.  samples_per_frame=8192

First 81920 samples of node 5 rawdump data from 2015-04-27 GMRT observations of
the Crab pulsar.
"""

SAMPLE_GSB_PHASED_HEADER = _full_path('gsb/sample_gsb_phased.timestamp')
"""GSB phased header sample.

10 header entries, starting from seq_nr=9994, from 2013-07-27 GMRT observations
of PSR J1810+1744.
"""

SAMPLE_GSB_PHASED = ((_full_path('gsb/sample_gsb_phased.Pol-L1.dat'),
                      _full_path('gsb/sample_gsb_phased.Pol-L2.dat')),
                     (_full_path('gsb/sample_gsb_phased.Pol-R1.dat'),
                      _full_path('gsb/sample_gsb_phased.Pol-R2.dat')))
"""GSB phased sample.  samples_per_frame=8

80 complete samples, starting from seq_nr=9994, from 2013-07-27 GMRT
observations of PSR J1810+1744, rewritten so each frame has 8 complete samples.
"""

SAMPLE_DRAO_CORRUPT = _full_path('sample_drao_corrupted.vdif')
"""Corrupted VDIF sample. bps=4.

First ten frames extracted from b0329 DRAO corrupted raw data file
0059000.dat.
"""



# Licensed under the GPLv3 - see LICENSE
import os

import numpy as np
import astropy.units as u
from astropy.time import Time

from .. import vdif
from ..helpers import sequentialfile
from ..helpers.tests.test_sequentialfile import Sequencer


def test_sequentialfile_vdif_stream(tmpdir):
    vdif_sequencer = Sequencer(str(tmpdir.join('{:07d}.vdif')))
    # Try writing a very simple file, using edv=0.
    data = np.ones((16, 16, 2, 2))
    for i, dat in enumerate(data):
        dat[i, 0, 0] = -1.
        dat[i, 1, 1] = -1.
    data.shape = -1, 2, 2
    # construct first header
    header = vdif.VDIFHeader.fromvalues(
        edv=0, time=Time('2010-01-01'), nchan=2, bps=2,
        complex_data=False, frame_nr=0, thread_id=0, samples_per_frame=16,
        station='me')
    with sequentialfile.open(vdif_sequencer, 'w+b',
                             file_size=4*header.frame_nbytes) as sfh, \
            vdif.open(sfh, 'ws', header0=header, nthread=2,
                      sample_rate=256*u.Hz) as fw:
        fw.write(data)
    # check that this wrote 8 frames
    files = [vdif_sequencer[i] for i in range(8)]
    for file_ in files:
        assert os.path.isfile(file_)
    assert not os.path.isfile(vdif_sequencer[8])

    with sequentialfile.open(vdif_sequencer, 'rb') as sfh, vdif.open(
            sfh, 'rs', sample_rate=256*u.Hz) as fr:
        record1 = fr.read(21)
        assert np.all(record1 == data[:21])
        fr.seek(7*16)
        record2 = fr.read(61)
        assert np.all(record2 == data[7*16:7*16+61])
        assert fr.tell() == 7*16+61
    # Might as well check file list too.
    with sequentialfile.open(files, 'rb') as sfh, vdif.open(
            sfh, 'rs', sample_rate=256*u.Hz) as fr:
        record1 = fr.read()
        assert np.all(record1 == data)



# Licensed under the GPLv3 - see LICENSE
import io

import pytest
import numpy as np
from astropy import units as u
from astropy.time import Time

from ... import vdif
from ...data import SAMPLE_VDIF as SAMPLE_FILE


class TestCorruptSampleCopy:
    @classmethod
    def setup_class(cls):
        # Make a triply-long sample file - this since otherwise
        # things already fail at the determination of thread_ids.
        with vdif.open(SAMPLE_FILE, 'rs') as fs, \
                io.BytesIO() as s, \
                vdif.open(s, 'ws', header0=fs.header0, nthread=8) as fw:

            data = fs.read()
            for i in range(3):
                fw.write(data)

            cls.data = np.concatenate([data, data, data])

            s.seek(0)
            cls.sample_bytes = s.read()
            cls.frame_nbytes = fw.header0.frame_nbytes
            cls.start_time = fw.start_time
            cls.stop_time = fw.tell('time')

    def test_sample_bytes(self, tmpdir):
        test_file = str(tmpdir.join('test.vdif'))
        with open(test_file, 'wb') as fh:
            fh.write(self.sample_bytes)
        with vdif.open(test_file, 'rs') as fs:
            data = fs.read()
        assert np.all(data == self.data)

    # Have 6 framesets, so 48 frames.
    @pytest.mark.parametrize('missing', (
        36, slice(46, 48), [30, 45], slice(8, 16), 0, slice(4, 12)))
    def test_missing_frames(self, missing, tmpdir):
        """Purely missing frames should just be marked invalid."""
        # Even at the very start; gh-359
        sample = np.frombuffer(self.sample_bytes, 'u1').reshape(-1, 5032)
        use = np.ones(len(sample), bool)
        use[missing] = False
        reduced = sample[use]
        corrupt_file = str(tmpdir.join('missing_frames.vdif'))
        with open(corrupt_file, 'wb') as s:
            s.write(reduced.tobytes())

        with vdif.open(corrupt_file, 'rb') as fr:
            assert 'number_of_frames' not in fr.info.warnings
            if np.count_nonzero(use) % 8 == 0:
                assert 'number_of_framesset' not in fr.info.warnings
            else:
                assert 'number_of_framesets' in fr.info.warnings

        with vdif.open(corrupt_file, 'rs') as fh:
            with pytest.warns(UserWarning,
                              match='problem loading frame'):
                data = fh.read()

        # Get data in frame order to zero expected bad frames.
        expected = (self.data.copy().reshape(-1, 20000, 8)
                    .transpose(0, 2, 1).reshape(-1, 20000))
        expected[missing] = 0.
        # Back into regular order
        expected = (expected.reshape(-1, 8, 20000)
                    .transpose(0, 2, 1).reshape(-1, 8))

        assert np.all(expected == data)

    def expected_bad_frames(self, missing):
        (start_f, start_r), (stop_f, stop_i) = [
            divmod(s, self.frame_nbytes)
            for s in (missing.start, missing.stop-1)]

        if start_r < 32 and start_f % 8 != 0:
            start_f -= 1

        return start_f, stop_f+1

    @pytest.mark.parametrize('missing,expected_bad_start,expected_bad_stop', [
        (slice(50320, 50321), 9, 11),  # First byte of header of frame 10.
        (slice(50500, 50600), 10, 11),  # Part of payload of frame 10.
        (slice(60000, 70000), 11, 14),  # Parts of 11-13.
        (slice(75490, 75500), 14, 16),  # Part of header of frame 15.
        (slice(80511, 80512), 15, 16)])  # Last byte of last frame.
    def test_expected_bad_frames(self, missing, expected_bad_start,
                                 expected_bad_stop):
        bad_start, bad_stop = self.expected_bad_frames(missing)
        assert bad_start == expected_bad_start
        assert bad_stop == expected_bad_stop

    # Keep frames in first three frame sets intact for get_thread_ids()
    @pytest.mark.parametrize('missing', [
        (slice(5032*26, 5032*26+1)),  # First byte of header of frame 26.
        (slice(5032*26+50, 5032*26+60)),  # Part of payload of frame 26.
        (slice(5032*27+50, 5032*29+700)),  # Parts of 27-29
        (slice(5032*31+10, 5032*31+20)),  # Part of header of frame 31.
        (slice(5032*48-1, 5032*48))])  # Last byte of last frame.
    def test_missing_bytes(self, missing, tmpdir):
        corrupted = (self.sample_bytes[:missing.start]
                     + self.sample_bytes[missing.stop:])
        bad_start, bad_stop = self.expected_bad_frames(missing)

        filename = str(tmpdir.join('corrupted.vdif'))
        with open(filename, 'wb') as fw:
            fw.write(corrupted)

        with vdif.open(filename, 'rb') as fr:
            assert 'number_of_frames' in fr.info.warnings

        # Check that bad frames are found with verify only.
        with vdif.open(filename, 'rs', verify=True) as fv:
            assert not fv.info.readable
            assert not fv.info.checks['continuous']
            assert 'continuous' in fv.info.errors
            # Reading will fail the frameset *before* the one tested.
            expected_msg = 'While reading at {}'.format(
                (bad_start // 8 - 1) * fv.samples_per_frame)
            assert expected_msg in fv.info.errors['continuous']

        # While only warnings are given when it is fixable.
        with vdif.open(filename, 'rs', verify='fix') as ff:
            assert ff.info.readable
            assert 'fixable' in ff.info.checks['continuous']
            assert 'continuous' in ff.info.warnings
            assert expected_msg in ff.info.warnings['continuous']
            assert 'problem loading frame' in ff.info.warnings['continuous']

        # Now check that data is properly marked as invalid.
        with vdif.open(filename, 'rs') as fr:
            assert fr.start_time == self.start_time
            assert fr.stop_time == self.stop_time
            with pytest.warns(UserWarning,
                              match='problem loading frame'):
                data = fr.read()

        # Get data in frame order to zero expected bad frames.
        expected = (self.data.copy().reshape(-1, 20000, 8)
                    .transpose(0, 2, 1).reshape(-1, 20000))
        expected[bad_start:bad_stop] = 0.
        # Back into regular order
        expected = (expected.reshape(-1, 8, 20000)
                    .transpose(0, 2, 1).reshape(-1, 8))

        assert np.all(data == expected)


class TestCorruptFile:
    @classmethod
    def setup_class(cls):
        cls.header0 = vdif.VDIFHeader.fromvalues(
            edv=1, time=Time('2010-11-12T13:14:15'), nchan=2, bps=2,
            complex_data=False, thread_id=0, samples_per_frame=16,
            station='me', sample_rate=2*u.kHz)
        cls.nthread = 2
        cls.data = np.array([[[-1, 1],
                              [-3, 3]]]*16)
        cls.frameset_nbytes = cls.header0.frame_nbytes * cls.nthread

    def fake_file(self, tmpdir, nframes=16):
        filename = str(tmpdir.join('fake.vdif'))
        with vdif.open(filename, 'ws', header0=self.header0,
                       nthread=self.nthread) as fw:
            for _ in range(nframes):
                fw.write(self.data)
        return filename

    def corrupt_copy(self, filename, missing):
        corrupt_name = filename.replace('.vdif', '_corrupt.vdif')
        with open(filename, 'rb') as fr, \
                open(corrupt_name, 'wb') as fw:
            fw.write(fr.read(missing.start))
            fr.seek(missing.stop)
            fw.write(fr.read())
        return corrupt_name

    @pytest.mark.parametrize('frame_nr', [1, 3, 5, slice(7, 10)])
    def test_missing_frameset(self, frame_nr, tmpdir):
        if not isinstance(frame_nr, slice):
            frame_nr = slice(frame_nr, frame_nr+1)
        missing = slice(frame_nr.start * self.frameset_nbytes,
                        frame_nr.stop * self.frameset_nbytes)
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, missing)
        with vdif.open(corrupt_file, 'rs') as fr:
            with pytest.warns(UserWarning, match='All threads'):
                data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        assert np.all(data[:frame_nr.start].astype(int) == self.data)
        assert np.all(data[frame_nr.stop:].astype(int) == self.data)
        assert np.all(data[frame_nr] == 0.)

    @pytest.mark.parametrize('frame_nr,thread', [
        (3, 0), (3, 1), (1, 1), (15, 1)])
    def test_missing_thread(self, frame_nr, thread, tmpdir):
        frame = frame_nr * self.nthread + thread
        missing = slice(frame * self.header0.frame_nbytes,
                        (frame+1) * self.header0.frame_nbytes)
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, missing)
        with vdif.open(corrupt_file, 'rs') as fr:
            with pytest.warns(UserWarning,
                              match='Thread.*{0}.*missing'.format(thread)):
                data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        assert np.all(data[:frame_nr].astype(int) == self.data)
        assert np.all(data[frame_nr+1:].astype(int) == self.data)
        assert np.all(data[frame_nr, :, thread] == 0.)
        assert np.all(data[frame_nr, :, 1-thread].astype(int)
                      == self.data[:, 1-thread])

    @pytest.mark.parametrize('missing_bytes', [
        slice(0, 80),  # Remove whole last frame set.
        slice(0, 40),  # Remove first thread of last frame
        slice(0, 32),  # Remove first header of last frame.
        slice(16, 32),  # Corrupt first header of last frame.
        slice(0, 16),  # Corrupt first header of last frame.
        slice(0, 1),  # Corrupt header byte of last frame.
        slice(10, 11),  # Corrupt header byte of last frame.
        slice(15, 16),  # Corrupt header byte of last frame.
        slice(20, 21),  # Corrupt header byte of last frame.
        slice(23, 24),  # Corrupt header byte of last frame.
    ])
    def test_missing_end(self, missing_bytes, tmpdir):
        # In all these cases, the data read should just be short.
        missing = slice(missing_bytes.start + 15*self.frameset_nbytes,
                        missing_bytes.stop + 15*self.frameset_nbytes)
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, missing)
        with vdif.open(corrupt_file, 'rs') as fr:
            assert fr.size == 15 * self.data.size
            data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        assert len(data) == 15
        assert np.all(data.astype(int) == self.data)

    # Note: keep frame sets 0--2 intact for get_thread_ids().
    @pytest.mark.parametrize('missing_bytes,missing_data', [
        (slice(240, 320), slice(48, 64)),  # Remove frameset 3.
        (slice(279, 281), slice(48, 64)),  # Corrupt frameset 3.
        (slice(280, 281), slice(48, 64)),  # Corrupt frameset 3, thread 1.
        (slice(279, 280), slice(48, 64)),  # Corrupt frameset 3, thread 0.
        (slice(272, 365), slice(48, 80)),  # Corrupt framesets 3, 4
    ])
    def test_missing_middle(self, missing_bytes, missing_data, tmpdir):
        # In all these cases, some data will be missing.
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, missing_bytes)
        with vdif.open(corrupt_file, 'rs') as fr:
            assert fr.size == 16 * self.data.size
            with pytest.warns(UserWarning, match='problem loading frame set'):
                data = fr.read()

        expected = np.concatenate([self.data] * 16)
        expected[missing_data] = 0.
        assert np.all(data.astype(int) == expected)

    def test_duplicate_data(self, tmpdir):
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, slice(320, 240))
        with vdif.open(corrupt_file, 'rs') as fv:
            with pytest.raises(Exception, match='excess data'):
                fv.read()

    def test_bad_thread_ids(self, tmpdir):
        filename = str(tmpdir.join('fake.vdif'))
        header = self.header0.copy()
        with vdif.open(filename, 'wb') as fw:
            for i in range(4):
                for j in range(i+1):
                    header['frame_nr'] = i
                    header['thread_id'] = j
                    fw.write_frame(self.data[:, 0], header)

        with vdif.open(filename, 'rb') as fr:
            assert len(fr.read_frameset().frames) == 1
            assert len(fr.read_frameset().frames) == 2
            assert len(fr.read_frameset().frames) == 3
            assert len(fr.read_frameset().frames) == 4

        with pytest.raises(EOFError):
            vdif.open(filename, 'rs')


class TestInvalidFrameHeaders:
    # CHIME VDIF files from ARO can have invalid frames included in
    # which the header information -- in particular the frame_nr and
    # seconds -- are corrupt as well.  Check that we can skip those.
    @classmethod
    def setup_class(cls):
        cls.header0 = vdif.VDIFHeader.fromvalues(
            edv=1, time=Time('2010-11-12T13:14:15'), nchan=2, bps=2,
            complex_data=False, thread_id=0, samples_per_frame=16,
            station='me', sample_rate=2*u.kHz)
        cls.nthread = 2
        cls.data = np.array([[[-1, 1],
                              [-3, 3]]]*16)
        cls.frameset_nbytes = cls.header0.frame_nbytes * cls.nthread

    def fake_file(self, tmpdir):
        filename = str(tmpdir.join('fake.vdif'))
        with vdif.open(filename, 'wb') as fw:
            for i in range(16):
                header = self.header0.copy()
                if i != 10:
                    header['frame_nr'] = i
                else:
                    header['frame_nr'] = 0
                    header['seconds'] = 0
                    header['invalid_data'] = True
                fw.write_frameset(self.data, header=header)

        return filename

    @pytest.mark.parametrize('verify', ('fix', False))
    def test_invalid_frame_fix(self, verify, tmpdir):
        fake_file = self.fake_file(tmpdir)
        with vdif.open(fake_file, 'rs', verify=verify) as fr:
            data = fr.read()

        expected = np.stack([self.data] * 16)
        expected[10] = 0
        expected.shape = (-1,) + self.data.shape[1:]
        assert np.all(data.astype(int) == expected)

    def test_invalid_frame_nofix(self, tmpdir):
        fake_file = self.fake_file(tmpdir)
        with pytest.raises(ValueError, match='wrong frame number'):
            with vdif.open(fake_file, 'rs', verify=True) as fr:
                fr.read()



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for GUPPI headers.

Implements a GUPPIHeader class that reads & writes FITS-like headers from file.
"""
import operator

import astropy.units as u
from astropy.io import fits
from astropy.time import Time, TimeDelta


__all__ = ['GUPPIHeader']


class GUPPIHeader(fits.Header):
    """GUPPI baseband file format header.

    Parameters
    ----------
    *args : str or iterable
        If a string, parsed as a GUPPI header from a file, otherwise
        as for the `astropy.io.fits.Header` baseclass.
    verify : bool, optional
        Whether to do minimal verification that the header is consistent with
        the GUPPI standard.  Default: `True`.
    mutable : bool, optional
        Whether to allow the header to be changed after initialisation.
        Default: `True`.
    **kwargs
        Any further header keywords to be set.

    Notes
    -----
    Like `~astropy.io.fits.Header`, the initialiser does not accept keyword
    arguments to populate an array - instead, one must pass an iterable. In
    order to ensure keywords are kept in the right order, one should pass on
    values as a tuple, not as a dict. E.g., to copy a header, one should not
    do ``GUPPIHeader({key: header[key] for key in header})``, but rather::

        GUPPIHeader(((key, header[key]) for key in header))

    or, to also keep the comments::

        GUPPIHeader(((key, (header[key], header.comments[key]))
                      for key in header))
    """

    _properties = ('payload_nbytes', 'frame_nbytes', 'bps', 'nchan', 'npol',
                   'sample_shape', 'sample_rate', 'sideband', 'overlap',
                   'samples_per_frame', 'offset', 'start_time',
                   'time')
    """Properties accessible/usable in initialisation for all headers."""

    _defaults = [('BACKEND', 'GUPPI'),
                 ('BLOCSIZE', 0),
                 ('STT_OFFS', 0),
                 ('PKTIDX', 0),
                 ('OVERLAP', 0),
                 ('SRC_NAME', 'unset'),
                 ('TELESCOP', 'unset'),
                 ('PKTFMT', '1SFA'),
                 ('PKTSIZE', 8192),
                 ('NBITS', 8),
                 ('NPOL', 1),
                 ('OBSNCHAN', 1)]

    supported_formats = {'1SFA', 'SIMPLE'}
    """GUPPI formats that are known to work.

    '1SFA' is used for all modes other than FAST4K (which is only used
    for total intensity).  'SIMPLE' is from DSPSR, and used to support
    time-first payloads.  See
    https://safe.nrao.edu/wiki/pub/Main/JoeBrandt/guppi_status_shmem.pdf

    If a format is not in this set, yet is known to work, a PR would be
    most welcome.
    """

    def __init__(self, *args, verify=True, mutable=True, **kwargs):
        # Comments handled by fits.Header__init__().
        super().__init__(*args, **kwargs)
        self.mutable = mutable
        # Empty header always OK, since things will be added to it.
        if len(self) and verify:
            self.verify()

    def verify(self):
        """Basic check of integrity."""
        # Same check as dspsr's dsp::GUPPIFile::is_valid
        assert all(key in self for key in ('BLOCSIZE',
                                           'PKTIDX'))
        # We could check here for self['PKTFMT'] in self.supported_formats
        # but that would break reading of unsupported but working formats,
        # so instead this becomes just a warning in file_info.

    def copy(self):
        """Create a mutable and independent copy of the header."""
        # This method exists because io.fits.Header.copy has a docstring that
        # refers to `Header` which breaks sphinx...
        return super().copy()

    @classmethod
    def fromfile(cls, fh, verify=True):
        """Reads in GUPPI header block from a file.

        Parameters
        ----------
        fh : filehandle
            To read data from.
        verify: bool, optional
            Whether to do basic checks on whether the header is valid. Verify
            is automatically called by `~astropy.io.fits.Header.fromstring`, so
            this flag exists only to standardize the API.
        """
        header_start = fh.tell()
        # Find the size of the header.  GUPPI header entries are 80 char long
        # with <=8 char keyword names.  "=" is always the 9th char.
        line = '========='
        while line[8] in ('=', ' '):
            line = fh.read(80).decode('ascii')
            if line[:3] == 'END':
                break
            if line == '':
                raise EOFError

        header_end = fh.tell()
        fh.seek(header_start)
        hdr = fh.read(header_end - header_start).decode('ascii')
        # Create the header using the base class.
        self = cls.fromstring(hdr)
        self.mutable = False
        if verify:
            self.verify()
        # GUPPI headers are not a proper FITS standard, and we're reading
        # from a file that the user likely cannot control, so let's not bother
        # with card verification (this avoids warnings in repr(); gh-282)
        for c in self.cards:
            c._verified = True
        return self

    def tofile(self, fh):
        """Write GUPPI file header to filehandle.

        Uses `~astropy.io.fits.Header.tostring`.
        """
        fh.write(self.tostring(padding=False).encode('ascii'))

    @classmethod
    def fromkeys(cls, *args, verify=True, mutable=True, **kwargs):
        """Initialise a header from keyword values.

        Like fromvalues, but without any interpretation of keywords.

        Note that this just passes kwargs to the class initializer as a dict
        (for compatibility with fits.Header). It is present for compatibility
        with other header classes only.
        """
        return cls(kwargs, *args, verify=verify, mutable=mutable)

    @classmethod
    def fromvalues(cls, **kwargs):
        """Initialise a header from parsed values.

        Here, the parsed values must be given as keyword arguments, i.e., for
        any ``header``, ``cls.fromvalues(**header) == header``.

        However, unlike for the ``fromkeys`` class method, data can also be set
        using arguments named after header methods, such as ``time``.

        Furthermore, some header defaults are set in ``GUPPIHeader._defaults``.
        """
        self = cls(cls._defaults, verify=False)
        self.update(**kwargs)
        return self

    def update(self, *, verify=True, **kwargs):
        """Update the header with new values.

        Here, any keywords matching properties are processed as well, in the
        order set by the class (in ``_properties``), and after all other
        keywords have been processed.

        Parameters
        ----------
        verify : bool, optional
            If `True` (default), verify integrity after updating.
        **kwargs
            Arguments used to set keywords and properties.
        """
        # Remove kwargs that set properties, in correct order.
        extras = [(key, kwargs.pop(key)) for key in self._properties
                  if key in kwargs]
        # Update the normal keywords.
        super().update(kwargs)
        # Now set the properties.
        for attr, value in extras:
            setattr(self, attr, value)
        if verify:
            self.verify()

    def __setitem__(self, key, value):
        if not self.mutable:
            raise TypeError("immutable {0} does not support assignment."
                            .format(type(self).__name__))

        super().__setitem__(key.upper(), value)

    @property
    def nbytes(self):
        """Size of the header in bytes."""
        return (len(self) + 1) * 80

    @property
    def payload_nbytes(self):
        """Size of the payload in bytes."""
        return self['BLOCSIZE']

    @payload_nbytes.setter
    def payload_nbytes(self, payloadsize):
        self['BLOCSIZE'] = payloadsize

    @property
    def frame_nbytes(self):
        """Size of the frame in bytes."""
        return self.nbytes + self.payload_nbytes

    @frame_nbytes.setter
    def frame_nbytes(self, frame_nbytes):
        self.payload_nbytes = frame_nbytes - self.nbytes

    @property
    def bps(self):
        """Bits per elementary sample."""
        return self['NBITS']

    @bps.setter
    def bps(self, bps):
        self['NBITS'] = bps

    @property
    def complex_data(self):
        """Whether the data are complex."""
        return self['OBSNCHAN'] != 1

    @property
    def npol(self):
        """Number of polarisations."""
        return self['NPOL'] // (2 if self.complex_data else 1)

    @npol.setter
    def npol(self, npol):
        self['NPOL'] = npol * (2 if self.complex_data else 1)

    @property
    def nchan(self):
        """Number of channels."""
        return self['OBSNCHAN']

    @nchan.setter
    def nchan(self, nchan):
        self['OBSNCHAN'] = operator.index(nchan)

    @property
    def sample_shape(self):
        """Shape of a sample in the payload (npol, nchan)."""
        return self.npol, self.nchan

    @sample_shape.setter
    def sample_shape(self, sample_shape):
        # Need to set nchan first to properly set npol, since nchan is
        # connected to complex_data.
        self.nchan = sample_shape[1]
        self.npol = sample_shape[0]

    @property
    def _bpcs(self):
        """Bits per complete sample."""
        # NPOL includes factor of 2 for real/complex components.
        return self['OBSNCHAN'] * self['NPOL'] * self.bps

    @property
    def sample_rate(self):
        """Number of complete samples per second.

        Can be set with a negative quantity to set `sideband`.  Overlap samples
        are not included in the rate.
        """
        return (1. / self['TBIN']) * u.Hz

    @sample_rate.setter
    def sample_rate(self, sample_rate):
        self['TBIN'] = 1. / abs(sample_rate.to_value(u.Hz))
        bw = (sample_rate.to_value(u.MHz) * self['OBSNCHAN']
              / (1 if self.complex_data else 2))
        self['OBSBW'] = bw

    @property
    def sideband(self):
        """True if upper sideband."""
        return self['OBSBW'] > 0

    @sideband.setter
    def sideband(self, sideband):
        self['OBSBW'] = (1 if sideband else -1) * abs(self['OBSBW'])

    @property
    def channels_first(self):
        """True if encoded payload ordering is (nchan, nsample, npol)."""
        # Called ``time-ordered`` in DSPSR.
        return self['PKTFMT'] != 'SIMPLE'

    @channels_first.setter
    def channels_first(self, channels_first):
        self['PKTFMT'] = '1SFA' if bool(channels_first) else 'SIMPLE'

    @property
    def samples_per_frame(self):
        """Number of complete samples in the frame, including overlap."""
        return self.payload_nbytes * 8 // self._bpcs

    @samples_per_frame.setter
    def samples_per_frame(self, samples_per_frame):
        old_payload_nbytes = self.payload_nbytes
        self.payload_nbytes = (samples_per_frame * self._bpcs + 7) // 8
        if self.samples_per_frame != samples_per_frame:
            exc = ValueError("header cannot store {} samples per frame. "
                             "Nearest is {}.".format(samples_per_frame,
                                                     self.samples_per_frame))
            self.payload_nbytes = old_payload_nbytes
            raise exc

    @property
    def overlap(self):
        """Number of complete samples that overlap with the next frame."""
        return self['OVERLAP']

    @overlap.setter
    def overlap(self, overlap):
        self['OVERLAP'] = operator.index(overlap)

    @property
    def offset(self):
        """Offset from start of observation in units of time."""
        # PKTIDX only counts valid packets, not overlap ones.
        return ((self['PKTIDX'] * self['PKTSIZE'] * 8 // self._bpcs)
                * self['TBIN']) * u.s

    @offset.setter
    def offset(self, offset):
        self['PKTIDX'] = int((offset / (self['TBIN'] * u.s) / self['PKTSIZE']
                              * ((self._bpcs + 7) // 8)).to(u.one).round())

    @property
    def start_time(self):
        """Start time of the observation."""
        return (Time(self['STT_IMJD'], scale='utc', format='mjd')
                + TimeDelta(self['STT_SMJD'], self['STT_OFFS'], format='sec'))

    @start_time.setter
    def start_time(self, start_time):
        start_time = Time(start_time, scale='utc')
        imjd = int(start_time.mjd)
        # Calculate differences from start of day.
        djd = start_time - Time(imjd, format='mjd', scale='utc')
        # Correct for possible rounding errors.
        imjd += int(djd.jd)
        # Get seconds.  Should now be guaranteed to be between 0 and 86400
        # (or 86401 if a leap-second day).
        seconds = (start_time - Time(imjd, format='mjd', scale='utc')).sec
        self['STT_IMJD'] = imjd
        self['STT_SMJD'], self['STT_OFFS'] = divmod(seconds, 1)

    @property
    def time(self):
        """Start time of the part of the observation covered by this header."""
        return self.start_time + self.offset

    @time.setter
    def time(self, time):
        """Set header time.

        If ``start_time`` is not already set, this sets it using the time and
        ``offset``.  Otherwise, this sets ``offset`` using the time and
        ``start_time``.

        Parameters
        ----------
        time : `~astropy.time.Time`
            Time for the first sample associated with this header.
        """
        if 'STT_IMJD' not in self.keys():
            self.start_time = time - self.offset
        else:
            self.offset = time - self.start_time

    def _ipython_key_completions_(self):
        # Enables tab-completion of header keys in IPython.
        return self.keys()

    def __eq__(self, other):
        """Whether headers have the same keys with the same values."""
        return all(self.get(k, None) == other.get(k, None)
                   for k in (set(self.keys()) | set(other.keys())))

    def __repr__(self):
        name = self.__class__.__name__
        vals = super().__repr__()
        return('<{0} {1}>'.format(name, ("\n  " + len(name) * " ").join(
            [v.rstrip() for v in vals.split('\n')])))



# Licensed under the GPLv3 - see LICENSE
from copy import copy, deepcopy
import io
import pickle
from collections import namedtuple

import pytest
import numpy as np
import astropy.units as u

from ..header import HeaderParser, VLBIHeaderBase, four_word_struct
from ..payload import PayloadBase
from ..frame import FrameBase
from ..base import (FileBase, VLBIFileReaderBase, StreamBase,
                    StreamReaderBase, StreamWriterBase)


def encode_1bit(values):
    return np.packbits(values.ravel())


def decode_1bit(values):
    return np.unpackbits(values.view(np.uint8)).astype(np.float32)


def encode_8bit(values):
    return np.clip(np.round(values),
                   -128, 127).astype(np.int8)


def decode_8bit(values):
    return values.view(np.int8).astype(np.float32)


class Payload(PayloadBase):
    _encoders = {1: encode_1bit,
                 8: encode_8bit}
    _decoders = {1: decode_1bit,
                 8: decode_8bit}


class TestBase:
    def setup_class(cls):
        cls.header_parser = HeaderParser(
            (('x0_16_4', (0, 16, 4)),
             ('x0_31_1', (0, 31, 1, False)),
             ('x1_0_32', (1, 0, 32)),
             ('x2_0_64', (2, 0, 64, 1 << 32))))

        class Header(VLBIHeaderBase):
            _struct = four_word_struct
            _header_parser = cls.header_parser
            payload_nbytes = 8
            bps = 8
            sample_shape = (2,)
            complex_data = False

        cls.Header = Header
        cls.header = cls.Header([0x12345678, 0xffff0000, 0x0, 0xffffffff])
        cls.Payload = Payload
        cls.payload = Payload(np.array([0x12345678, 0xffff0000],
                                       dtype=Payload._dtype_word),
                              sample_shape=(2,), bps=8, complex_data=False)
        cls.payload1bit = Payload(np.array([0x12345678]*5,
                                           dtype=Payload._dtype_word),
                                  sample_shape=(5,), bps=1, complex_data=True)

        class Frame(FrameBase):
            _header_class = Header
            _payload_class = Payload

        cls.Frame = Frame
        cls.frame = Frame(cls.header, cls.payload)

    def test_header_basics(self):
        header = self.Header(None)
        assert header.words == [0] * 4
        with pytest.raises(Exception):
            self.Header([1] * 5)
        with pytest.raises(Exception):
            self.Header([1] * 3)
        header = self.header.copy()
        assert header == self.header
        assert header is not self.header
        header = copy(self.header)
        assert header == self.header
        assert header is not self.header
        assert type(header.words) is list
        header.mutable = False
        assert type(header.words) is tuple
        header = self.Header(0, verify=False)
        with pytest.raises(Exception):
            header.verify()
        with pytest.raises(TypeError):
            header.mutable = True

    def test_header_fromfile(self, tmpdir):
        with open(str(tmpdir.join('test.dat')), 'w+b') as s:
            s.write(four_word_struct.pack(*self.header.words))
            s.seek(2)
            with pytest.raises(EOFError):
                self.Header.fromfile(s)
            s.seek(0)
            header = self.Header.fromfile(s)
        assert header == self.header

    def test_parser(self):
        """Test that parsers work as expected."""
        assert self.header['x0_16_4'] == 4
        assert self.header['x0_31_1'] is False
        assert self.header['x1_0_32'] == self.header.words[1]
        assert (self.header['x2_0_64']
                == self.header.words[2] + self.header.words[3] * (1 << 32))
        assert 'x0_31_1' in self.header
        assert 'bla' not in self.header
        with pytest.raises(KeyError):
            self.header['bla']
        with pytest.raises(KeyError):
            self.header['bla'] = 1
        with pytest.raises(AttributeError):
            self.header.x0_16_4
        with pytest.raises(AttributeError):
            self.header.xbla

    def test_make_setter(self):
        header = self.header.copy()
        header['x0_16_4'] = 0xe
        assert header.words[0] == 0x123e5678
        header['x0_16_4'] = 0
        assert header.words[0] == 0x12305678
        header['x0_16_4'] = True  # Special value: fill all bits
        assert header.words[0] == 0x123f5678
        with pytest.raises(ValueError,
                           match='cannot be represented with 4 bits'):
            header['x0_16_4'] = 0x10
        with pytest.raises(ValueError, match='no default value'):
            header['x0_16_4'] = None

        header['x0_31_1'] = True
        assert header.words[0] == 0x923f5678
        header['x1_0_32'] = 0x1234
        assert header.words[:2] == [0x923f5678, 0x1234]
        header['x2_0_64'] = 1
        assert header.words[2:] == [1, 0]
        header['x2_0_64'] = True  # fill all bits
        assert header.words[2:] == [0xffffffff, 0xffffffff]
        header['x2_0_64'] = None  # set to default
        assert header.words[2:] == [0, 1]
        # Also check update method.
        header.update(x1_0_32=0x5678, x2_0_64=1)
        assert header.words == [0x923f5678, 0x5678, 1, 0]
        with pytest.warns(UserWarning, match='unused.*bla'):
            header.update(bla=10)
        with pytest.raises(ValueError, match='cannot be represented'):
            header.update(x0_16_4=-1)

    def test_header_without_invariants(self):
        # More elaborate tests implicitly done in locate_frames.
        assert self.Header.invariants() == set()
        assert self.header.invariants() == set()
        with pytest.raises(ValueError):
            self.header.invariant_pattern()
        with pytest.raises(ValueError):
            self.Header.invariant_pattern()

        # But we can construct one by passing on the invariants ourselves,
        # if those have defaults.
        pattern, mask = self.Header.invariant_pattern({'x0_31_1'})
        assert pattern == [0, 0, 0, 0]
        assert mask == [0x80000000, 0, 0, 0]
        # If there is no default, we cannot use a key on the class.
        with pytest.raises(ValueError):
            self.Header.invariant_pattern({'x1_0_32'})
        # But we can on the instance.
        pattern, mask = self.header.invariant_pattern({'x1_0_32'})
        assert pattern == self.header.words
        assert mask == [0, 0xffffffff, 0, 0]

    def test_header_with_implicit_invariants(self):
        class SyncHeader(self.Header):
            _header_parser = self.Header._header_parser | HeaderParser(
                (('sync_pattern', (1, 0, 32, 0x12345678)),))

        assert SyncHeader.invariants() == {'sync_pattern'}
        pattern, mask = SyncHeader.invariant_pattern()
        assert pattern == [0, 0x12345678, 0, 0]
        assert mask == [0, 0xffffffff, 0, 0]
        sync_header = SyncHeader.fromvalues(x2_0_64=10)
        assert sync_header.invariants() == {'sync_pattern'}
        pattern, mask = sync_header.invariant_pattern()
        assert pattern == [0, 0x12345678, 10, 0]
        assert mask == [0x00000000, 0xffffffff, 0, 0]
        pattern, mask = sync_header.invariant_pattern({'x0_16_4',
                                                       'sync_pattern'})
        assert pattern == [0, 0x12345678, 10, 0]
        assert mask == [0xf0000, 0xffffffff, 0, 0]
        # If given keys have no default, the class version should fail.
        with pytest.raises(ValueError):
            SyncHeader.invariant_pattern({'x0_16_4', 'sync_pattern'})

    def test_header_with_explicit_invariants(self):
        # On purpose, check the old way of "adding" HeaderParser,
        # instead of following python 3.9 and using "|" for update,
        # to be sure we support code written against baseband < 4.0.
        class SyncHeader(self.Header):
            _header_parser = self.Header._header_parser + HeaderParser(
                (('sync_pattern', (1, 0, 32, 0x12345678)),))
            _invariants = {'sync_pattern'}
            _stream_invariants = _invariants | {'x0_31_1'}

        assert SyncHeader.invariants() == {'sync_pattern'}
        pattern, mask = SyncHeader.invariant_pattern()
        assert pattern == [0, 0x12345678, 0, 0]
        assert mask == [0, 0xffffffff, 0, 0]
        sync_header = SyncHeader.fromvalues(x2_0_64=10)
        assert sync_header.invariants() == {'sync_pattern', 'x0_31_1'}
        pattern, mask = sync_header.invariant_pattern()
        assert pattern == [0, 0x12345678, 10, 0]
        assert mask == [0x80000000, 0xffffffff, 0, 0]
        pattern, mask = sync_header.invariant_pattern({'x0_16_4',
                                                       'sync_pattern'})
        assert pattern == [0, 0x12345678, 10, 0]
        assert mask == [0xf0000, 0xffffffff, 0, 0]
        # If given keys have no default, the class version should fail.
        with pytest.raises(ValueError):
            SyncHeader.invariant_pattern({'x0_16_4', 'sync_pattern'})

    def test_payload_basics(self):
        assert self.payload.complex_data is False
        assert self.payload.sample_shape == (2,)
        assert self.payload.bps == 8
        assert self.payload.nbytes == 8
        assert self.payload.shape == (4, 2)
        assert self.payload.size == 8
        assert self.payload.ndim == 2
        assert np.all(self.payload.data.ravel()
                      == self.payload.words.view(np.int8))
        assert np.all(np.array(self.payload).ravel()
                      == self.payload.words.view(np.int8))
        assert np.all(np.array(self.payload, dtype=np.int8).ravel()
                      == self.payload.words.view(np.int8))
        payload = self.Payload(self.payload.words, bps=4)
        with pytest.raises(KeyError):
            payload.data
        with pytest.raises(ValueError):
            self.Payload(self.payload.words.astype('>u4'), bps=4)
        payload = self.Payload(self.payload.words, bps=8, complex_data=True)
        assert np.all(payload.data
                      == (self.payload.data[:, 0]
                          + 1j * self.payload.data[:, 1]))

        assert self.payload1bit.complex_data is True
        assert self.payload1bit.sample_shape == (5,)
        assert self.payload1bit.bps == 1
        assert self.payload1bit.shape == (16, 5)
        assert self.payload1bit.nbytes == 20
        assert np.all(self.payload1bit.data.ravel()
                      == (np.unpackbits(self.payload1bit.words.view(np.uint8))
                          .astype(np.float32).view(np.complex64)))

    @pytest.mark.parametrize('item', (2, slice(1, 3), (), slice(2, None),
                                      (2, 1), (slice(None), 0),
                                      (slice(1, 3), 1)))
    def test_payload_getitem_setitem(self, item):
        data = self.payload.data
        sel_data = data[item]
        assert np.all(self.payload[item] == sel_data)
        payload = self.Payload(self.payload.words.copy(), sample_shape=(2,),
                               bps=8, complex_data=False)
        assert payload == self.payload
        payload[item] = 1 - sel_data
        check = self.payload.data
        check[item] = 1 - sel_data
        assert np.all(payload[item] == 1 - sel_data)
        assert np.all(payload.data == check)
        assert np.all(payload[:]
                      == payload.words.view(np.int8).reshape(-1, 2))
        assert payload != self.payload
        payload[item] = sel_data
        assert np.all(payload[item] == sel_data)
        assert payload == self.payload
        payload = self.Payload.fromdata(data + 1j * data, bps=8)
        sel_data = payload.data[item]
        assert np.all(payload[item] == sel_data)
        payload[item] = 1 - sel_data
        check = payload.data
        check[item] = 1 - sel_data
        assert np.all(payload.data == check)

    def test_payload_bad_fbps(self):
        with pytest.raises(TypeError):
            self.payload1bit[10:11]

    def test_payload_empty_item(self):
        p11 = self.payload[1:1]
        assert p11.size == 0
        assert p11.shape == (0,) + self.payload.sample_shape
        assert p11.dtype == self.payload.dtype
        payload = self.Payload(self.payload.words.copy(), sample_shape=(2,),
                               bps=8, complex_data=False)
        payload[1:1] = 1
        assert payload == self.payload

    @pytest.mark.parametrize('item', (20, -20, (slice(None), 5)))
    def test_payload_invalid_item(self, item):
        with pytest.raises(IndexError):
            self.payload[item]

        payload = self.Payload(self.payload.words.copy(), sample_shape=(2,),
                               bps=8, complex_data=False)
        with pytest.raises(IndexError):
            payload[item] = 1

    def test_payload_invalid_item2(self):
        with pytest.raises(TypeError):
            self.payload['l']
        payload = self.Payload(self.payload.words.copy(), sample_shape=(2,),
                               bps=8, complex_data=False)
        with pytest.raises(TypeError):
            payload['l'] = 1

    def test_payload_setitem_wrong_shape(self):
        payload = self.Payload(self.payload.words.copy(), sample_shape=(2,),
                               bps=8, complex_data=False)
        with pytest.raises(ValueError):
            payload[1] = np.ones(10)

        with pytest.raises(ValueError):
            payload[1] = np.ones((2, 2))

        with pytest.raises(ValueError):
            payload[1:3] = np.ones((2, 3))

        with pytest.raises(ValueError):
            payload[1:3, 0] = np.ones((2, 2))

        with pytest.raises(ValueError):
            payload[1:3, :1] = np.ones((1, 2))

    def test_payload_fixed_size(self):
        class FixedSizePayload(self.Payload):
            _nbytes = self.Header.payload_nbytes

        fsp = FixedSizePayload(self.payload.words, header=self.header)
        assert fsp.nbytes == 8

        FixedSizePayload._nbytes = 6
        with pytest.raises(ValueError, match='header payload size should'):
            FixedSizePayload(self.payload.words, header=self.header)
        with pytest.raises(ValueError, match='encoded data should'):
            FixedSizePayload(self.payload.words)

    def test_payload_fromfile(self, tmpdir):
        with open(str(tmpdir.join('test.dat')), 'w+b') as s:
            self.payload.tofile(s)
            s.seek(0)
            with pytest.raises(ValueError):
                self.Payload.fromfile(s)  # No size given.
            s.seek(0)
            payload = self.Payload.fromfile(
                s, payload_nbytes=len(self.payload.words) * 4,
                sample_shape=(2,), bps=8)
        assert payload == self.payload

    def test_payload_fromdata(self):
        data = np.random.normal(0., 64., 16).reshape(16, 1)
        payload = self.Payload.fromdata(data, bps=8)
        assert payload.complex_data is False
        assert payload.sample_shape == (1,)
        assert payload.bps == 8
        assert payload.words.dtype is self.Payload._dtype_word
        assert len(payload.words) == 4
        assert len(payload) == len(data)
        assert payload.nbytes == 16
        payload2 = self.Payload.fromdata(self.payload.data,
                                         bps=self.payload.bps)
        assert payload2 == self.payload
        header = self.header.copy()
        header.bps = 8
        payload3 = self.Payload.fromdata(self.payload.data,
                                         header=header)
        assert payload3 == self.payload
        with pytest.raises(ValueError, match='data are complex'):
            self.Payload.fromdata(self.payload.data.astype(complex), header)
        payload4 = self.Payload.fromdata(data.ravel(), bps=8)
        assert payload4.sample_shape == ()
        assert payload4.shape == (16,)
        assert payload4 != payload
        assert np.all(payload4.data == payload.data.ravel())
        with pytest.raises(ValueError):  # don't have relevant encoder.
            self.Payload.fromdata(data, bps=4)
        payload5 = self.Payload.fromdata(data[::2, 0] + 1j * data[1::2, 0],
                                         bps=8)
        assert payload5.complex_data is True
        assert payload5.sample_shape == ()
        assert payload5.shape == (8,)
        assert payload5 != payload
        assert np.all(payload5.words == payload.words)

    def test_frame_basics(self):
        assert self.frame.header is self.header
        assert self.frame.payload is self.payload
        assert len(self.frame) == len(self.payload)
        assert self.frame.sample_shape == self.payload.sample_shape
        assert self.frame.shape == self.payload.shape
        assert self.frame.size == self.payload.size
        assert self.frame.ndim == self.payload.ndim
        assert np.all(self.frame.data == self.payload.data)
        assert np.all(np.array(self.frame) == np.array(self.payload))
        assert np.all(np.array(self.frame, dtype=np.float64)
                      == np.array(self.payload))
        assert self.frame.valid is True
        frame = self.Frame(self.header, self.payload, valid=False)
        assert np.all(frame.data == 0.)
        frame.fill_value = 1.
        assert np.all(frame.data == 1.)

        assert 'x2_0_64' in self.frame
        assert self.frame['x2_0_64'] == self.header['x2_0_64']
        for item in (3, (1, 1), slice(0, 2)):
            assert np.all(self.frame[item] == self.frame.payload[item])

        frame2 = self.Frame(self.header.copy(),
                            self.Payload.fromdata(self.payload.data,
                                                  bps=self.payload.bps))

        assert frame2.header == self.header
        assert frame2.payload == self.payload
        assert frame2 == self.frame
        frame2['x2_0_64'] = 0x1
        assert frame2['x2_0_64'] == 0x1
        assert frame2.header != self.header
        frame2[3, 1] = 5.
        assert frame2[3, 1] == 5
        assert frame2.payload != self.payload

    def test_frame_fromfile(self, tmpdir):
        with open(str(tmpdir.join('test.dat')), 'w+b') as s:
            self.frame.tofile(s)
            s.seek(0)
            frame = self.Frame.fromfile(s, payload_nbytes=self.payload.nbytes,
                                        sample_shape=(2,), bps=8)
        assert frame == self.frame

    def test_frame_fromdata(self):
        frame = self.Frame.fromdata(self.frame.data, self.header, bps=8)
        assert frame == self.frame
        frame2 = self.Frame.fromdata(self.frame.data, self.header,
                                     bps=8, valid=False)
        assert np.all(frame2.data == 0.)

    def test_vlbi_file_base(self, tmpdir):
        # This is probably too basic to show that the wrapper works.
        filename = str(tmpdir.join('test.dat'))
        with io.open(filename, 'wb') as fw:
            fh = FileBase(fw)
            assert fh.fh_raw is fw
            assert not fh.readable()
            assert fh.writable()
            assert not fh.closed
            with pytest.raises(AttributeError):
                fh.bla
            assert repr(fh).startswith('FileBase(fh_raw')
            fh.write(b'abcd')
            fh.close()
            assert fh.closed
            assert fh.fh_raw.closed
        with io.open(filename, 'rb') as fr:
            fh = FileBase(fr)
            assert fh.fh_raw is fr
            assert fh.readable()
            assert not fh.writable()
            assert not fh.closed
            with pytest.raises(AttributeError):
                fh.bla
            assert repr(fh).startswith('FileBase(fh_raw')
            assert fh.read() == b'abcd'
            fh.close()
            assert fh.closed
            assert fh.fh_raw.closed

    def test_temporary_offset(self, tmpdir):
        filename = str(tmpdir.join('test.dat'))
        with io.open(filename, 'wb') as fw:
            fw.write(b'abcdefghijklmnopqrstuvwxyz')

        with io.open(filename, 'rb') as fr:
            fh = FileBase(fr)
            fh.seek(2)
            assert fh.read(2) == b'cd'
            assert fh.tell() == 4
            with fh.temporary_offset():
                assert fh.seek(1) == 1
                assert fh.read(2) == b'bc'
                assert fh.tell() == 3
            assert fh.tell() == 4
            with fh.temporary_offset(-2, 2):
                assert fh.read(1) == b'y'
                assert fh.tell() == 25
            assert fh.tell() == 4

    def test_pickle(self, tmpdir):
        filename = str(tmpdir.join('test.dat'))
        with io.open(filename, 'wb') as fw:
            fw.write(b'abcdefghijklmnopqrstuvwxyz')

        with VLBIFileReaderBase(io.open(filename, 'rb')) as fh:
            assert fh.read(2) == b'ab'
            pickled = pickle.dumps(fh)
            with pickle.loads(pickled) as fh2:
                assert fh2.tell() == 2
                assert fh2.read(2) == b'cd'
                fh2.seek(-2, 2)
                assert fh2.read(2) == b'yz'

            # cannot read closed file
            with pytest.raises(ValueError):
                fh2.read()

            assert fh.tell() == 2
            assert fh.read(2) == b'cd'

        with pickle.loads(pickled) as fh3:
            assert fh3.read(2) == b'cd'

        # cannot read closed file
        with pytest.raises(ValueError):
            fh3.read()

        closed = pickle.dumps(fh)
        with pickle.loads(closed) as fh4:
            assert fh4.closed
            with pytest.raises(ValueError):
                fh4.read()

        with FileBase(io.open(filename, 'wb')) as fw:
            with pytest.raises(TypeError):
                pickle.dumps(fw)

    def test_deepcopy(self, tmpdir):
        filename = str(tmpdir.join('test.dat'))
        with io.open(filename, 'wb') as fw:
            fw.write(b'abcdefghijklmnopqrstuvwxyz')

        with VLBIFileReaderBase(io.open(filename, 'rb')) as fh:
            assert fh.read(2) == b'ab'
            with deepcopy(fh) as fh2:
                assert fh2.tell() == 2
                assert fh2.read(2) == b'cd'
                fh2.seek(-2, 2)
                assert fh2.read(2) == b'yz'

            # cannot read closed file
            with pytest.raises(ValueError):
                fh2.read()

            # Old file was not moved or closed.
            assert fh.tell() == 2
            assert fh.read(2) == b'cd'

        with deepcopy(fh) as fh3:
            assert fh3.closed
            # cannot read closed file
            with pytest.raises(ValueError):
                fh3.read()

        with FileBase(io.open(filename, 'wb')) as fw:
            with pytest.raises(TypeError):
                deepcopy(fw)


class TestSqueezeAndSubset:
    def setup(self):
        self.sample_shape_maker = namedtuple('SampleShape',
                                             'n0, n1, n2, n3, n4')
        self.unsliced_shape = (1, 21, 33, 1, 2)
        self.squeezed_shape = (21, 33, 2)
        self.squeezed_fields = ('n1', 'n2', 'n4')
        self.unsliced_data = np.ones((100,) + self.unsliced_shape, dtype='f4')
        self.squeezed_data = self.unsliced_data.squeeze()
        self.header_class = namedtuple('Header',
                                       'sample_rate,samples_per_frame,'
                                       'sample_shape,bps,complex_data')
        self.other_args = dict(bps=1, complex_data=False,
                               samples_per_frame=1000, sample_rate=10000*u.Hz)
        self.header0 = self.header_class(sample_shape=self.unsliced_shape,
                                         **self.other_args)

    def make_reader_with_shape(self, squeeze=True, subset=None,
                               sample_shape_maker=None, unsliced_shape=None):

        class StreamReaderWithShape(StreamReaderBase):
            _sample_shape_maker = sample_shape_maker or self.sample_shape_maker

        header0 = self.header_class(
            sample_shape=unsliced_shape or self.unsliced_shape,
            **self.other_args)

        return StreamReaderWithShape(
            None, header0=header0, squeeze=squeeze, subset=subset)

    def make_writer_with_shape(self, squeeze=True, sample_shape_maker=None,
                               unsliced_shape=None):

        class StreamWriterWithShape(StreamWriterBase):
            _sample_shape_maker = sample_shape_maker or self.sample_shape_maker

        header0 = self.header_class(
            sample_shape=unsliced_shape or self.unsliced_shape,
            **self.other_args)

        return StreamWriterWithShape(None, header0=header0, squeeze=squeeze)

    def test_sample_shape_and_squeeze(self):
        # Tests stream base's sample and squeezing routines.
        # Try tuple only.
        sb = StreamBase(fh_raw=None, header0=self.header0, squeeze=False)
        assert sb.sample_shape == self.unsliced_shape
        sb = StreamBase(fh_raw=None, header0=self.header0, squeeze=True)
        assert sb.sample_shape == self.squeezed_shape

        # Try reader with equivalent sample shape.
        sr = self.make_reader_with_shape(squeeze=False)
        assert sr.sample_shape == self.unsliced_shape
        assert sr.sample_shape._fields == self.sample_shape_maker._fields

        sr = self.make_reader_with_shape(squeeze=True)
        assert sr.sample_shape == self.squeezed_shape
        assert sr.sample_shape._fields == self.squeezed_fields
        assert (sr._squeeze_and_subset(self.unsliced_data).shape
                == self.squeezed_data.shape)
        assert (sr._squeeze_and_subset(self.unsliced_data[:1]).shape
                == (1,) + self.squeezed_shape)

        # With StreamWriterBase, we can access _unsqueeze.
        sw = self.make_writer_with_shape(squeeze=False)
        assert sw.sample_shape == self.unsliced_shape
        assert sw.sample_shape._fields == self.sample_shape_maker._fields

        sw = self.make_writer_with_shape(squeeze=True)
        assert sw.sample_shape == self.squeezed_shape
        assert sw.sample_shape._fields == self.squeezed_fields
        assert (sw._unsqueeze(self.squeezed_data).shape
                == self.unsliced_data.shape)
        assert (sw._unsqueeze(self.squeezed_data[:1]).shape
                == (1,) + self.unsliced_shape)

        # Check that single-axis sample shape squeezes to ().
        sample_shape_maker_s = namedtuple('SampleShape', 'n0')
        unsliced_shape_short = (1,)
        sws = self.make_writer_with_shape(False, sample_shape_maker_s,
                                          unsliced_shape_short)
        assert sws.sample_shape == (1,)
        sws = self.make_writer_with_shape(True, sample_shape_maker_s,
                                          unsliced_shape_short)
        assert sws.sample_shape == ()
        data = np.empty(100, dtype='float32')
        assert sws._unsqueeze(data).shape == (100, 1)

    @pytest.mark.parametrize(
        ('squeeze', 'subset', 'sliced_shape', 'sliced_n'),
        [(False, None, (1, 21, 33, 1, 2), ('n0', 'n1', 'n2', 'n3', 'n4')),
         (True, None, (21, 33, 2), ('n1', 'n2', 'n4')),
         (False, (), (1, 21, 33, 1, 2), ('n0', 'n1', 'n2', 'n3', 'n4')),
         (True, (), (21, 33, 2), ('n1', 'n2', 'n4')),
         (False, 0, (21, 33, 1, 2), ('n1', 'n2', 'n3', 'n4')),
         (True, 0, (33, 2), ('n2', 'n4')),
         (False, (0, 13), (33, 1, 2), ('n2', 'n3', 'n4')),
         (True, (0, 13), (2,), ('n4',)),
         (False, (Ellipsis, 0, 1), (1, 21, 33), None),
         (True, (Ellipsis, 0, 1), (21,), ('n1',)),
         (False, (0, slice(1, None, 4), slice(None), 0),
          (5, 33, 2), ('n1', 'n2', 'n4')),
         (True, (slice(1, None, 4), slice(None), [1]),
          (5, 33, 1), ('n1', 'n2', 'n4')),
         (False, (0, 0, slice(None, 1, -4)), (8, 1, 2), ('n2', 'n3', 'n4')),
         (True, (0, slice(None, 1, -4)), (8, 2), ('n2', 'n4')),
         (False, (0, np.array([2, 8, 9])[:, np.newaxis], [1, 7]),
          (3, 2, 1, 2), None),
         (True, (np.array([2, 8, 9])[:, np.newaxis], [1, 7], 0),
          (3, 2), None)])
    def test_squeeze_subset(self, squeeze, subset, sliced_shape, sliced_n):
        # Tests subsetting for squeezed samples.
        sb = self.make_reader_with_shape(squeeze=squeeze, subset=subset)
        assert sb.squeeze == squeeze
        if isinstance(subset, tuple):
            assert sb.subset == subset
        elif subset is None:
            assert sb.subset == ()
        else:
            assert sb.subset == (subset,)
        assert sb.sample_shape == sliced_shape
        assert getattr(sb.sample_shape, '_fields', None) == sliced_n
        subset_data = sb._squeeze_and_subset(self.unsliced_data)
        assert subset_data.shape == (100,) + sliced_shape

    def test_faulty_subset(self):
        # Advanced indexing changes dimensions, so sample_shape can't be set.
        with pytest.raises(IndexError) as excinfo:
            self.make_reader_with_shape(
                subset=([0], np.array([2, 8, 16])[:, np.newaxis], [1, 7]))
        assert "cannot be used to" in str(excinfo.value)

        # Can't subset with a string.
        with pytest.raises(IndexError) as excinfo:
            self.make_reader_with_shape(subset=(0, 'nonsense', [1, 7]))
        assert "cannot be used to" in str(excinfo.value)

        # Numerical index 8 is out of bounds of 3rd dimension
        with pytest.raises(IndexError) as excinfo:
            self.make_reader_with_shape(subset=(3, 0, [2, 8]))
        assert "cannot be used to" in str(excinfo.value)

        # Slice is out of bounds of 3rd dimension.
        with pytest.raises(AssertionError) as excinfo:
            self.make_reader_with_shape(subset=(3, 0, slice(4, 8)))



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for VLBI Mark 5B payloads.

Implements a Mark5BPayload class used to store payload words, and decode to
or encode from a data array.

For the specification, see
https://www.haystack.mit.edu/tech/vlbi/mark5/docs/Mark%205B%20users%20manual.pdf
"""
from collections import namedtuple

import numpy as np

from ..base.payload import PayloadBase
from ..base.encoding import encode_2bit_base, decoder_levels
from ..base.utils import fixedvalue


__all__ = ['init_luts', 'decode_1bit', 'decode_2bit',
           'encode_1bit', 'encode_2bit',
           'Mark5BPayload']


# Some duplication with mark4.py here: lut2bit = mark4.lut2bit1
# Though lut1bit = -mark4.lut1bit, so perhaps not worth combining.
def init_luts():
    """Set up the look-up tables for levels as a function of input byte.

    For 1-bit mode, one has just the sign bit:
     === =====
      s  value
     === =====
      0  +1
      1  -1
     === =====

    For 2-bit mode, there is a sign and a magnitude, which encode:
     === === ===== =====
      m   s  value s*2+m
     === === ===== =====
      0   0  -Hi    0
      0   1  +1     2
      1   0  -1     1
      1   1  +Hi    3
     === === ===== =====

    Note that the sign bit is flipped for 1-bit mode from what one might
    expect from the 2-bit encodings (at least, as implemented in mark5access;
    the docs are rather unclear).  For more details, see Table 13 in
    https://library.nrao.edu/public/memos/vlba/up/VLBASU_13.pdf
    and
    https://www.haystack.mit.edu/tech/vlbi/mark5/docs/Mark%205B%20users%20manual.pdf
    Appendix A: sign always on even bit stream (0, 2, 4, ...), and magnitude
    on adjacent odd stream (1, 3, 5, ...).

    In the above table, the last column is the index in the linearly increasing
    table of levels (``decoder_levels[2]``).
    """
    b = np.arange(256)[:, np.newaxis]
    sl = np.arange(8)
    # For 1-bit modes, like for Mark 4, if set, sign=-1, so need to get item 0.
    # With this, one reproduces mark5access m5d, but must say that from the
    # documentation, it is not obvious this is correct.
    lut1bit = decoder_levels[1][((b >> sl) & 1) ^ 1]
    # 2-bit mode: sign bit in lower position thatn magnitude bit
    # ms=00,01,10,11 = -Hi, 1, -1, Hi (lut
    s = np.arange(0, 8, 2)  # 0, 2, 4, 6
    m = s + 1               # 1, 3, 5, 7
    sl = (((b >> s) & 1) << 1) + ((b >> m) & 1)
    lut2bit = decoder_levels[2][sl]
    return lut1bit, lut2bit


lut1bit, lut2bit = init_luts()


def decode_1bit(words):
    b = words.view(np.uint8)
    return lut1bit.take(b, axis=0)


shift1bit = np.arange(0, 8).astype(np.uint8)


def encode_1bit(values):
    """Encodes values using 1 bit per sample, packing the result into bytes."""
    bitvalues = np.signbit(values.reshape(-1, 8)).view(np.uint8)
    return np.packbits(bitvalues[:, ::-1])


def decode_2bit(words):
    b = words.view(np.uint8)
    return lut2bit.take(b, axis=0)


shift2bit = np.arange(0, 8, 2).astype(np.uint8)
reorder = np.array([0, 2, 1, 3], dtype=np.uint8)


def encode_2bit(values):
    bitvalues = encode_2bit_base(values.reshape(-1, 4))
    # swap 1 & 2
    reorder.take(bitvalues, out=bitvalues)
    bitvalues <<= shift2bit
    return np.bitwise_or.reduce(bitvalues, axis=-1)


encode_2bit.__doc__ = encode_2bit_base.__doc__


class Mark5BPayload(PayloadBase):
    """Container for decoding and encoding VDIF payloads.

    Parameters
    ----------
    words : `~numpy.ndarray`
        Array containg LSB unsigned words (with the right size) that
        encode the payload.
    nchan : int, optional
        Number of channels.   Default: 1.
    bps : int, optional
        Bits per elementary sample.  Default: 2.
    """

    _nbytes = 2500 * 4
    _encoders = {1: encode_1bit,
                 2: encode_2bit}
    _decoders = {1: decode_1bit,
                 2: decode_2bit}

    _sample_shape_maker = namedtuple('SampleShape', 'nchan')

    @fixedvalue
    def complex_data(self):
        return False

    @classmethod
    def fromdata(cls, data, *, bps=2):
        """Encode data as payload, using a given number of bits per sample.

        It is assumed that the last dimension is the number of channels.
        """
        # Override just to remove option of passing in header.
        return super().fromdata(data, bps=bps)



# Licensed under the GPLv3 - see LICENSE
import pytest
import numpy as np
from numpy.testing import assert_array_equal

from ..utils import lcm, bcd_encode, bcd_decode, byte_array, CRC, CRCStack


class TestBCD:
    def test_bcd_decode(self):
        assert bcd_decode(0x1) == 1
        assert bcd_decode(0x9123) == 9123
        with pytest.raises(ValueError):
            bcd_decode(0xf)
        decoded = bcd_decode(np.array([0x1, 0x9123]))
        assert isinstance(decoded, np.ndarray)
        assert np.all(decoded == np.array([1, 9123]))
        with pytest.raises(ValueError):
            bcd_decode(np.array([0xf, 9123]))
        with pytest.raises(TypeError):
            bcd_decode([1, 2])

    def test_bcd_encode(self):
        assert bcd_encode(1) == 0x1
        assert bcd_encode(9123) == 0x9123
        with pytest.raises(TypeError):
            bcd_encode('bla')

    def test_roundtrip(self):
        assert bcd_decode(bcd_encode(15)) == 15
        assert bcd_decode(bcd_encode(8765)) == 8765
        a = np.array([1, 9123])
        assert np.all(bcd_decode(bcd_encode(a)) == a)


class TestCRC12:

    @staticmethod
    def hex_to_stream(string):
        n = len(string) * 4
        scalar = int(string, base=16)
        return np.array([((scalar & (1 << bit)) != 0)
                         for bit in range(n-1, -1, -1)], bool)

    @staticmethod
    def stream_to_scalar(stream):
        n = len(stream)
        result = 0
        for i, s in enumerate(stream):
            result += int(s) << (n - i - 1)

        return result

    def setup(self):
        # Test example from page 4 of
        # https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf
        stream_hex = '0000 002D 0330 0000' + 'FFFF FFFF' + '4053 2143 3805 5'
        self.crc_hex = '284'
        self.crc12 = CRC(0x180f)
        self.crcstack12 = CRCStack(0x180f)

        self.stream_hex = stream_hex.replace(' ', '').lower()

        self.stream = int(self.stream_hex, base=16)
        self.bitstream = self.hex_to_stream(self.stream_hex)

        self.crc = int(self.crc_hex, base=16)
        self.crcstream = self.hex_to_stream(self.crc_hex)

    def test_setup(self):
        assert '{:037x}'.format(self.stream) == self.stream_hex
        assert '{:03x}'.format(self.crc) == self.crc_hex
        assert self.stream_to_scalar(self.bitstream) == self.stream
        assert self.stream_to_scalar(self.crcstream) == self.crc

    def test_crc_stream(self):
        crcstream = self.crcstack12(self.bitstream)
        assert np.all(crcstream == self.crcstream)

    def test_check_crc_stream(self):
        fullstream = np.hstack((self.bitstream, self.crcstream))
        assert self.crcstack12.check(fullstream)

    def test_crc_scalar(self):
        crc = self.crc12(self.stream)
        assert crc == self.crc

    def test_check_crc_scalar(self):
        scalar = (self.stream << len(self.crc12)) + self.crc
        assert self.crc12.check(scalar)

    def test_crc_array(self):
        scalar = 0x12345678
        expected = self.crc12(scalar)
        array = (scalar * np.ones(10, dtype=int)).astype('u8')
        crc = self.crc12(array)
        assert crc.shape == array.shape
        assert np.all(crc == expected)

    def test_check_crc_array(self):
        scalar = 0x12345678
        scalar = (scalar << len(self.crc12)) + self.crc12(scalar)
        array = (scalar * np.ones(10, dtype=int)).astype('u8')
        check = self.crc12.check(array)
        assert check.shape == array.shape
        assert np.all(check)


@pytest.mark.parametrize(
    ('a', 'b', 'lcm_out'),
    ((7, 14, 14),
     (7853, 6199, 48680747),
     (0, 5, 0),
     (4, -12, 12),
     (-4, -12, 12)))
def test_lcm(a, b, lcm_out):
    assert lcm(a, b) == lcm_out


@pytest.mark.parametrize(
    ('pattern', 'expected'),
    [(b'\xa0\x55', [160, 85]),
     (0x55a0, [160, 85, 0, 0]),
     ([0x55, 0xa0], [85, 0, 0, 0, 160, 0, 0, 0]),
     (np.array([0xa0, 0x55], 'u1'), [160, 85]),
     (np.array(0x55a0, '<u4'), [160, 85, 0, 0]),
     (np.array(0x55a0, '<u8'), [160, 85, 0, 0, 0, 0, 0, 0]),
     (np.array([0x55, 0xa0], '<u4'), [85, 0, 0, 0, 160, 0, 0, 0])])
def test_byte_array(pattern, expected):
    result = byte_array(pattern)
    expected = np.array(expected, 'u1')
    assert_array_equal(result, expected)


def test_byte_array_errors():
    with pytest.raises(ValueError, match='values have to fit'):
        byte_array([-1, -1])



# Licensed under the GPLv3 - see LICENSE
"""Baseband general I/O routines and entry point.

Contains general ``open`` and ``file_info`` functions that can iterate
over possible formats to determine which is the right one, including
possible formats discovered via entry point 'baseband.io', such as
the `~baseband_tasks.io.hdf5` module from the baseband-tasks_ package.

Any 'baseband.io' entry points are treated as possible formats if they
point to a module (e.g., 'vdif = baseband.vdif').  Any entries that
have object names (e.g., 'fancy_open = mypackage.io:open') are just
added to the name space.

Attributes
----------
FORMATS : list
    Available baseband formats.

"""
import sys


__all__ = ['open', 'file_info']


__self__ = sys.modules[__name__]
"""Link to our own module, for convenience below."""

# We only load entries on demand, to keep import time minimal.
_entries = {}
"""Entry points found."""
_bad_entries = set()
"""Any entry points that failed to load. These will not be retried."""


def __getattr__(attr):
    """Get a missing attribute from a possible entry point.

    Looks for the attribute among the (possibly updated) entry points,
    and, if found, tries loading the entry.  If that fails, the entry
    is added to _bad_entries to ensure it does not recur.
    """
    if sys.version_info >= (3, 8):
        from importlib.metadata import EntryPoint, entry_points
    else:
        from importlib_metadata import EntryPoint, entry_points

    if attr.startswith('_') or attr in _bad_entries:
        raise AttributeError(f"module {__name__!r} has no attribute {attr!r}")

    FORMATS = globals().setdefault('FORMATS', [])
    if attr not in _entries:
        if not _entries:
            # On initial update, we add our own formats as explicit entries,
            # in part to set some order, but also so things work even in a
            # pure source checkout, where entry points are missing.
            _entries.update({
                fmt: EntryPoint(fmt, 'baseband.'+fmt, '')
                for fmt in ('dada', 'guppi', 'mark4', 'mark5b', 'vdif', 'gsb')
            })

        # Note: again do not presume the entry points exist, since we may
        # be in a pure source checkout.
        _entries.update({entry_point.name: entry_point for entry_point
                         in entry_points().get('baseband.io', [])})
        # Need python >= 3.9 to be able to do entry.attr.
        FORMATS.extend([name for name, entry in _entries.items()
                        if not (entry.value.partition(':')[2]
                                or name in FORMATS)])
        if attr == 'FORMATS':
            return FORMATS

    entry = _entries.get(attr, None)
    if entry is None:
        raise AttributeError(f"module {__name__!r} has no attribute {attr!r}")

    try:
        value = entry.load()
    except Exception:
        _entries.pop(attr)
        _bad_entries.add(attr)
        if attr in FORMATS:
            FORMATS.remove(attr)
        raise AttributeError(f"{entry} was not loadable. Now removed")

    # Update so we do not have to go through __getattr__ again.
    globals()[attr] = value
    return value


def __dir__():
    # Force update of entries, creates 'FORMATS' if it doesn't exist.
    hasattr(__self__, 'absolutely_no_way_this_exists')
    return sorted(set(globals()).union(_entries).difference(_bad_entries))


def file_info(name, format=None, **kwargs):
    """Get format and other information from a baseband file.

    The keyword arguments will only be used if needed, so if one is unsure
    what format a file is, but knows it was taken recently and has 8 channels,
    one would put in ``ref_time=Time('2015-01-01'), nchan=8``. Alternatively,
    and perhaps easier, one can first call the function without extra arguments
    in which case the result will describe what is missing.

    Parameters
    ----------
    name : str or filehandle, or sequence of str
        Raw file for which to obtain information.  If a sequence of files is
        passed, returns information from the first file (see Notes).
    format : str, tuple of str, optional
        Formats to try.  If not given, try all standard formats.
    **kwargs
        Any arguments that might help to get information.  For instance,
        Mark 4 and Mark 5B do not have complete timestamps, which can be
        addressed by passing in ``ref_time``.  Furthermore, for Mark 5B, it
        is needed to pass in ``nchan``. Arguments are checked for consistency
        with the file even if not used (see notes below).

    Returns
    -------
    info
        The information on the file.  Depending on how much information could
        be gathered, this will be an instance of either
        `~baseband.base.file_info.StreamReaderInfo`,
        `~baseband.base.file_info.FileReaderInfo`, or
        `~baseband.base.file_info.NoInfo`.

    Notes
    -----
    All keyword arguments passed in are classified, ending up in one of
    the following (mostly useful if the file could be opened as a stream):

      - ``used_kwargs``: arguments that were needed to open the file.
      - ``consistent_kwargs``: not needed to open the file, but consistent.
      - ``inconsistent_kwargs``: not needed to open the file, and inconsistent.
      - ``irrelevant_kwargs``: provide information irrelevant for opening.
    """
    from baseband.base.file_info import NoInfo

    # If we're looking at one file but multiple formats, cycle through formats.
    if format is None:
        format = tuple(__self__.FORMATS)

    if isinstance(format, tuple):
        no_info = set()
        for format_ in format:
            info = file_info(name, format_, **kwargs)
            if info:
                return info

            if isinstance(info, NoInfo):
                no_info.add(format_)

        return NoInfo(f"{name} does not seem formatted as any of {set(format)}"
                      + (f" ({no_info} had no 'info' and opening failed)."
                         if no_info else "."))

    module = getattr(__self__, format)
    # A well-behaved module should define info, but we allow for more
    # minimally defined ones by trying to open the file and getting info
    # from it (which may well work if VLBIStreamReaderBase is subclasses).
    if hasattr(module, 'info'):
        return module.info(name, **kwargs)

    # Try just opening as stream and getting info.
    try:
        with module.open(name, 'rs', **kwargs) as fh:
            return fh.info
    except Exception as exc:
        return NoInfo(f"baseband.io.{format} has no 'info' and opening "
                      f"raised {exc!r}.")


def open(name, mode='rs', format=None, **kwargs):
    """Open a baseband file (or sequence of files) for reading or writing.

    Opened as a binary file, one gets a wrapped filehandle that adds
    methods to read/write a frame.  Opened as a stream, the handle is
    wrapped further, and reading and writing to the file is done as if
    the file were a stream of samples.

    Parameters
    ----------
    name : str or filehandle, or sequence of str
        File name, filehandle, or sequence of file names.  A sequence may be a
        list or str of ordered filenames, or an instance of
        `~baseband.helpers.sequentialfile.FileNameSequencer`.
    mode : {'rb', 'wb', 'rs', or 'ws'}, optional
        Whether to open for reading or writing, and as a regular binary
        file or as a stream. Default: 'rs', for reading a stream.
    format : str or tuple of str
        The format the file is in. For reading, if a tuple of possible formats,
        all will be tried in turn. By default, all supported formats are tried.
        For writing, an explicit format must be passed in.
    **kwargs
        Additional arguments needed for opening the file as a stream.
        For most formats, trying without these will raise an exception that
        tells which arguments are needed. Opening will not succeed if any
        arguments are passed in that are inconsistent with the file, or are
        irrelevant for opening the file.
    """
    if format is None or isinstance(format, tuple):
        if 'w' in mode:
            raise ValueError("cannot specify multiple formats for writing.")

        info = file_info(name, format, **kwargs)
        if not info:
            raise ValueError("format of file could not be auto-determined")

        format = info.format

        if getattr(info, 'missing', False) and 's' in mode:
            raise TypeError(f"file format {format} is missing required "
                            f"arguments {info.missing}.")

        if getattr(info, 'inconsistent_kwargs', False):
            raise ValueError(f"arguments inconsistent with this {format} file "
                             f"were passed in: {info.inconsistent_kwargs}")

        kwargs = getattr(info, 'used_kwargs', kwargs)
        # We add any arguments deemed irrelevant for opening the file, since
        # either they might just give some extra information (say, squeeze)
        # or they will just lead to a TypeError, in likely a more useful place.
        kwargs.update(getattr(info, 'irrelevant_kwargs', {}))

    module = getattr(__self__, format)
    return module.open(name, mode=mode, **kwargs)



# Licensed under the GPLv3 - see LICENSE
from ..base.frame import FrameBase
from .header import DADAHeader
from .payload import DADAPayload


__all__ = ['DADAFrame']


class DADAFrame(FrameBase):
    """Representation of a DADA file, consisting of a header and payload.

    Parameters
    ----------
    header : `~baseband.dada.DADAHeader`
        Wrapper around the header lines, providing access to the values.
    payload : `~baseband.dada.DADAPayload`
        Wrapper around the payload, provding mechanisms to decode it.
    valid : bool, optional
        Whether the data are valid.  Default: `True`.
    verify : bool, optional
        Whether to do basic verification of integrity.  Default: `True`.

    Notes
    -----
    DADA files do not support storing whether data are valid or not on disk.
    Hence, this has to be determined independently.  If ``valid=False``, any
    decoded data are set to ``cls.fill_value`` (by default, 0).

    The Frame can also be instantiated using class methods:

      fromfile : read header and map or read payload from a filehandle

      fromdata : encode data as payload

    Of course, one can also do the opposite:

      tofile : method to write header and payload to filehandle

      data : property that yields full decoded payload

    One can decode part of the payload by indexing or slicing the frame.

    A number of properties are defined: `shape`, `dtype` and `size` are
    the shape, type and number of complete samples of the data array, and
    `nbytes` the frame size in bytes.  Furthermore, the frame acts as a
    dictionary, with keys those of the header.  Any attribute that is not
    defined on the frame itself, such as ``.time`` will be looked up on the
    header as well.
    """
    _header_class = DADAHeader
    _payload_class = DADAPayload
